# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/type.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import predicate_pb2 as sarus__data__spec_dot_protobuf_dot_predicate__pb2
from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#sarus_data_spec/protobuf/type.proto\x12\x0fsarus_data_spec\x1a(sarus_data_spec/protobuf/predicate.proto\x1a#sarus_data_spec/protobuf/path.proto\"\xfa\x18\n\x04Type\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\x04null\x18\x03 \x01(\x0b\x32\x1a.sarus_data_spec.Type.NullH\x00\x12*\n\x04unit\x18\x04 \x01(\x0b\x32\x1a.sarus_data_spec.Type.UnitH\x00\x12\x30\n\x07\x62oolean\x18\x05 \x01(\x0b\x32\x1d.sarus_data_spec.Type.BooleanH\x00\x12\x30\n\x07integer\x18\x06 \x01(\x0b\x32\x1d.sarus_data_spec.Type.IntegerH\x00\x12*\n\x04\x65num\x18\x07 \x01(\x0b\x32\x1a.sarus_data_spec.Type.EnumH\x00\x12,\n\x05\x66loat\x18\x08 \x01(\x0b\x32\x1b.sarus_data_spec.Type.FloatH\x00\x12*\n\x04text\x18\t \x01(\x0b\x32\x1a.sarus_data_spec.Type.TextH\x00\x12,\n\x05\x62ytes\x18\n \x01(\x0b\x32\x1b.sarus_data_spec.Type.BytesH\x00\x12.\n\x06struct\x18\x0b \x01(\x0b\x32\x1c.sarus_data_spec.Type.StructH\x00\x12,\n\x05union\x18\x0c \x01(\x0b\x32\x1b.sarus_data_spec.Type.UnionH\x00\x12\x32\n\x08optional\x18\r \x01(\x0b\x32\x1e.sarus_data_spec.Type.OptionalH\x00\x12*\n\x04list\x18\x0e \x01(\x0b\x32\x1a.sarus_data_spec.Type.ListH\x00\x12,\n\x05\x61rray\x18\x0f \x01(\x0b\x32\x1b.sarus_data_spec.Type.ArrayH\x00\x12\x32\n\x08\x64\x61tetime\x18\x10 \x01(\x0b\x32\x1e.sarus_data_spec.Type.DatetimeH\x00\x12\x38\n\x0b\x63onstrained\x18\x11 \x01(\x0b\x32!.sarus_data_spec.Type.ConstrainedH\x00\x12\x36\n\nhypothesis\x18\x12 \x01(\x0b\x32 .sarus_data_spec.Type.HypothesisH\x00\x12&\n\x02id\x18\x13 \x01(\x0b\x32\x18.sarus_data_spec.Type.IdH\x00\x12*\n\x04\x64\x61te\x18\x14 \x01(\x0b\x32\x1a.sarus_data_spec.Type.DateH\x00\x12*\n\x04time\x18\x15 \x01(\x0b\x32\x1a.sarus_data_spec.Type.TimeH\x00\x12\x32\n\x08\x64uration\x18\x16 \x01(\x0b\x32\x1e.sarus_data_spec.Type.DurationH\x00\x12\x39\n\nproperties\x18\x02 \x03(\x0b\x32%.sarus_data_spec.Type.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x06\n\x04Null\x1a\x06\n\x04Unit\x1a\t\n\x07\x42oolean\x1a\xd0\x01\n\x07Integer\x12\x30\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32\".sarus_data_spec.Type.Integer.Base\x12\x0b\n\x03min\x18\x02 \x01(\x03\x12\x0b\n\x03max\x18\x03 \x01(\x03\x12\x17\n\x0fpossible_values\x18\x04 \x03(\x03\"`\n\x04\x42\x61se\x12\t\n\x05INT64\x10\x00\x12\t\n\x05INT32\x10\x01\x12\t\n\x05INT16\x10\x02\x12\x08\n\x04INT8\x10\x03\x12\n\n\x06UINT64\x10\x04\x12\n\n\x06UINT32\x10\x05\x12\n\n\x06UINT16\x10\x06\x12\t\n\x05UINT8\x10\x07\x1a\xde\x01\n\x04\x45num\x12-\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32\x1f.sarus_data_spec.Type.Enum.Base\x12\x0f\n\x07ordered\x18\x02 \x01(\x08\x12\x39\n\x0bname_values\x18\x03 \x03(\x0b\x32$.sarus_data_spec.Type.Enum.NameValue\x1a(\n\tNameValue\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03\"1\n\x04\x42\x61se\x12\t\n\x05INT64\x10\x00\x12\t\n\x05INT32\x10\x01\x12\t\n\x05INT16\x10\x02\x12\x08\n\x04INT8\x10\x03\x1a\x99\x01\n\x05\x46loat\x12.\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32 .sarus_data_spec.Type.Float.Base\x12\x0b\n\x03min\x18\x02 \x01(\x01\x12\x0b\n\x03max\x18\x03 \x01(\x01\x12\x17\n\x0fpossible_values\x18\x04 \x03(\x01\"-\n\x04\x42\x61se\x12\x0b\n\x07\x46LOAT64\x10\x00\x12\x0b\n\x07\x46LOAT32\x10\x01\x12\x0b\n\x07\x46LOAT16\x10\x02\x1a\x31\n\x04Text\x12\x10\n\x08\x65ncoding\x18\x01 \x01(\t\x12\x17\n\x0fpossible_values\x18\x02 \x03(\t\x1a\x07\n\x05\x42ytes\x1ax\n\x06Struct\x12\x32\n\x06\x66ields\x18\x01 \x03(\x0b\x32\".sarus_data_spec.Type.Struct.Field\x1a:\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12#\n\x04type\x18\x02 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1av\n\x05Union\x12\x31\n\x06\x66ields\x18\x01 \x03(\x0b\x32!.sarus_data_spec.Type.Union.Field\x1a:\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12#\n\x04type\x18\x02 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a/\n\x08Optional\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a=\n\x04List\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12\x10\n\x08max_size\x18\x02 \x01(\x03\x1a;\n\x05\x41rray\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12\r\n\x05shape\x18\x02 \x03(\x03\x1a\xb0\x01\n\x08\x44\x61tetime\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x0b\n\x03min\x18\x02 \x01(\t\x12\x0b\n\x03max\x18\x03 \x01(\t\x12\x17\n\x0fpossible_values\x18\x04 \x03(\t\x12\x31\n\x04\x62\x61se\x18\x05 \x01(\x0e\x32#.sarus_data_spec.Type.Datetime.Base\".\n\x04\x42\x61se\x12\x0c\n\x08INT64_NS\x10\x00\x12\x0c\n\x08INT64_MS\x10\x01\x12\n\n\x06STRING\x10\x02\x1a\x97\x01\n\x04\x44\x61te\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x0b\n\x03min\x18\x02 \x01(\t\x12\x0b\n\x03max\x18\x03 \x01(\t\x12\x17\n\x0fpossible_values\x18\x04 \x03(\t\x12-\n\x04\x62\x61se\x18\x05 \x01(\x0e\x32\x1f.sarus_data_spec.Type.Date.Base\"\x1d\n\x04\x42\x61se\x12\t\n\x05INT32\x10\x00\x12\n\n\x06STRING\x10\x01\x1a\xb6\x01\n\x04Time\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x0b\n\x03min\x18\x02 \x01(\t\x12\x0b\n\x03max\x18\x03 \x01(\t\x12\x17\n\x0fpossible_values\x18\x04 \x03(\t\x12-\n\x04\x62\x61se\x18\x05 \x01(\x0e\x32\x1f.sarus_data_spec.Type.Time.Base\"<\n\x04\x42\x61se\x12\x0c\n\x08INT64_NS\x10\x00\x12\x0c\n\x08INT32_MS\x10\x01\x12\n\n\x06STRING\x10\x02\x12\x0c\n\x08INT64_US\x10\x03\x1aK\n\x08\x44uration\x12\x0c\n\x04unit\x18\x01 \x01(\t\x12\x0b\n\x03min\x18\x02 \x01(\x03\x12\x0b\n\x03max\x18\x03 \x01(\x03\x12\x17\n\x0fpossible_values\x18\x04 \x03(\x03\x1a\x62\n\x0b\x43onstrained\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12.\n\nconstraint\x18\x02 \x01(\x0b\x32\x1a.sarus_data_spec.Predicate\x1a\x82\x01\n\nHypothesis\x12\x36\n\x05types\x18\x02 \x03(\x0b\x32\'.sarus_data_spec.Type.Hypothesis.Scored\x1a<\n\x06Scored\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12\r\n\x05score\x18\x02 \x01(\x01\x1a\xb5\x01\n\x02Id\x12+\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32\x1d.sarus_data_spec.Type.Id.Base\x12\x0e\n\x06unique\x18\x02 \x01(\x08\x12(\n\treference\x18\x03 \x01(\x0b\x32\x15.sarus_data_spec.Path\"H\n\x04\x42\x61se\x12\t\n\x05INT64\x10\x00\x12\t\n\x05INT32\x10\x01\x12\t\n\x05INT16\x10\x02\x12\x08\n\x04INT8\x10\x03\x12\n\n\x06STRING\x10\x04\x12\t\n\x05\x42YTES\x10\x05\x42\x06\n\x04typeb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.type_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TYPE_PROPERTIESENTRY._options = None
  _TYPE_PROPERTIESENTRY._serialized_options = b'8\001'
  _TYPE._serialized_start=136
  _TYPE._serialized_end=3330
  _TYPE_PROPERTIESENTRY._serialized_start=1167
  _TYPE_PROPERTIESENTRY._serialized_end=1216
  _TYPE_NULL._serialized_start=1218
  _TYPE_NULL._serialized_end=1224
  _TYPE_UNIT._serialized_start=1226
  _TYPE_UNIT._serialized_end=1232
  _TYPE_BOOLEAN._serialized_start=1234
  _TYPE_BOOLEAN._serialized_end=1243
  _TYPE_INTEGER._serialized_start=1246
  _TYPE_INTEGER._serialized_end=1454
  _TYPE_INTEGER_BASE._serialized_start=1358
  _TYPE_INTEGER_BASE._serialized_end=1454
  _TYPE_ENUM._serialized_start=1457
  _TYPE_ENUM._serialized_end=1679
  _TYPE_ENUM_NAMEVALUE._serialized_start=1588
  _TYPE_ENUM_NAMEVALUE._serialized_end=1628
  _TYPE_ENUM_BASE._serialized_start=1358
  _TYPE_ENUM_BASE._serialized_end=1407
  _TYPE_FLOAT._serialized_start=1682
  _TYPE_FLOAT._serialized_end=1835
  _TYPE_FLOAT_BASE._serialized_start=1790
  _TYPE_FLOAT_BASE._serialized_end=1835
  _TYPE_TEXT._serialized_start=1837
  _TYPE_TEXT._serialized_end=1886
  _TYPE_BYTES._serialized_start=1888
  _TYPE_BYTES._serialized_end=1895
  _TYPE_STRUCT._serialized_start=1897
  _TYPE_STRUCT._serialized_end=2017
  _TYPE_STRUCT_FIELD._serialized_start=1959
  _TYPE_STRUCT_FIELD._serialized_end=2017
  _TYPE_UNION._serialized_start=2019
  _TYPE_UNION._serialized_end=2137
  _TYPE_UNION_FIELD._serialized_start=1959
  _TYPE_UNION_FIELD._serialized_end=2017
  _TYPE_OPTIONAL._serialized_start=2139
  _TYPE_OPTIONAL._serialized_end=2186
  _TYPE_LIST._serialized_start=2188
  _TYPE_LIST._serialized_end=2249
  _TYPE_ARRAY._serialized_start=2251
  _TYPE_ARRAY._serialized_end=2310
  _TYPE_DATETIME._serialized_start=2313
  _TYPE_DATETIME._serialized_end=2489
  _TYPE_DATETIME_BASE._serialized_start=2443
  _TYPE_DATETIME_BASE._serialized_end=2489
  _TYPE_DATE._serialized_start=2492
  _TYPE_DATE._serialized_end=2643
  _TYPE_DATE_BASE._serialized_start=2614
  _TYPE_DATE_BASE._serialized_end=2643
  _TYPE_TIME._serialized_start=2646
  _TYPE_TIME._serialized_end=2828
  _TYPE_TIME_BASE._serialized_start=2768
  _TYPE_TIME_BASE._serialized_end=2828
  _TYPE_DURATION._serialized_start=2830
  _TYPE_DURATION._serialized_end=2905
  _TYPE_CONSTRAINED._serialized_start=2907
  _TYPE_CONSTRAINED._serialized_end=3005
  _TYPE_HYPOTHESIS._serialized_start=3008
  _TYPE_HYPOTHESIS._serialized_end=3138
  _TYPE_HYPOTHESIS_SCORED._serialized_start=3078
  _TYPE_HYPOTHESIS_SCORED._serialized_end=3138
  _TYPE_ID._serialized_start=3141
  _TYPE_ID._serialized_end=3322
  _TYPE_ID_BASE._serialized_start=3250
  _TYPE_ID_BASE._serialized_end=3322
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import (
    Collection,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)
import datetime
import typing as t

from sarus_data_spec.base import Referring
from sarus_data_spec.transform import Transform
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Scalar(Referring[sp.Scalar]):
    """A python class to describe scalars"""

    def __init__(self, protobuf: sp.Scalar, store: bool = True) -> None:
        if protobuf.spec.HasField("transformed"):
            transformed = protobuf.spec.transformed
            self._referred = {
                transformed.transform,
                *transformed.arguments,
                *list(transformed.named_arguments.values()),
            }

        super().__init__(protobuf=protobuf, store=store)

    def prototype(self) -> Type[sp.Scalar]:
        """Return the type of the underlying protobuf."""
        return sp.Scalar

    def name(self) -> str:
        return self._protobuf.name

    def doc(self) -> str:
        return self._protobuf.doc

    def spec(self) -> str:
        return str(self._protobuf.spec.WhichOneof('spec'))

    def is_transformed(self) -> bool:
        """Is the scalar composed."""
        return self._protobuf.spec.HasField("transformed")

    def is_remote(self) -> bool:
        """Is the dataspec a remotely defined dataset."""
        return self.manager().is_remote(self)

    def is_source(self) -> bool:
        """Is the scalar not composed."""
        return not self.is_transformed()

    def is_privacy_params(self) -> bool:
        """Is the scalar privacy parameters."""
        return self._protobuf.spec.HasField("privacy_params")

    def is_random_seed(self) -> bool:
        """Is the scalar a random seed."""
        if self._protobuf.spec.HasField("random_seed"):
            return True
        if self.is_transformed():
            transform = self.transform()
            if transform.protobuf().spec.HasField("derive_seed"):
                return True
        return False

    def is_synthetic_model(self) -> bool:
        """Is the scalar composed."""
        return self._protobuf.spec.HasField("synthetic_model")

    def is_pep(self) -> bool:
        """Is the scalar PEP."""
        return False

    def pep_token(self) -> Optional[str]:
        """Returns the scalar PEP token."""
        return None

    def rewritten_pep_token(self) -> Optional[str]:
        return None

    def is_synthetic(self) -> bool:
        """Is the scalar synthetic."""
        return self.manager().dataspec_validator().is_synthetic(self)

    def is_dp(self) -> bool:
        """Is the dataspec the result of a DP transform"""
        return self.manager().dataspec_validator().is_dp(self)

    def is_public(self) -> bool:
        """Is the scalar public."""
        return self.manager().dataspec_validator().is_public(self)

    def is_dp_writable(self) -> bool:
        """Check if it exists a variant of this dataspec
        that utilizes the dp equivalent, and this variant is is_dp."""
        return self.manager().dataspec_validator().is_dp_writable(self)

    def is_pep_writable(self) -> bool:
        """Check if it exists a variant of this dataspec that is is_pep."""
        return False

    def is_publishable(self) -> bool:
        """Check if it exists a variant of this dataspec that is published"""
        return self.manager().dataspec_validator().is_publishable(self)

    def is_published(self) -> bool:
        """Check if the dataspec is the result of a DP transform or another
        published dataspec.
        There is at least one parent that is DP.
        Such a dataspec has at least one private query in its computation
        graph."""
        return self.manager().dataspec_validator().is_published(self)

    def status(self, task_names: t.Optional[List[str]]) -> Optional[st.Status]:
        """This method return a status that contains all the
        last updates for the task_names required. It returns None if
        all the tasks are missing."""

        if task_names is None:
            task_names = []
        if type(task_names) not in [list, set, tuple]:
            raise TypeError(
                f"Invalid task_names passed to dataset.status {task_names}"
            )
        last_status = self.manager().status(self)
        if last_status is None:
            return last_status
        if all([last_status.task(task) is None for task in task_names]):
            return None
        return last_status

    def transform(self) -> st.Transform:
        return cast(
            st.Transform,
            self.storage().referrable(
                self.protobuf().spec.transformed.transform
            ),
        )

    def parents(
        self,
    ) -> Tuple[
        List[Union[st.DataSpec, st.Transform]],
        Dict[str, Union[st.DataSpec, st.Transform]],
    ]:
        if not self.is_transformed():
            return list(), dict()

        args_id = self._protobuf.spec.transformed.arguments
        kwargs_id = self._protobuf.spec.transformed.named_arguments

        args_parents = [
            cast(
                Union[st.DataSpec, st.Transform],
                self.storage().referrable(uuid),
            )
            for uuid in args_id
        ]
        kwargs_parents = {
            name: cast(
                Union[st.DataSpec, st.Transform],
                self.storage().referrable(uuid),
            )
            for name, uuid in kwargs_id.items()
        }

        return args_parents, kwargs_parents

    def parents_list(
        self,
    ) -> t.List[st.DataSpec]:
        parents_args, parents_kwargs = self.parents()
        parents_args.extend(parents_kwargs.values())

        return [
            dataspec_parent
            for dataspec_parent in parents_args
            if isinstance(dataspec_parent, st.DataSpec)
        ]

    def sources(
        self, type_name: t.Optional[str] = sp.type_name(sp.Dataset)
    ) -> Set[st.DataSpec]:
        """Returns the set of non-transformed datasets that are parents
        of the current dataset"""
        sources = self.storage().sources(self, type_name=type_name)
        return sources

    def variant(
        self,
        kind: st.ConstraintKind,
        public_context: Collection[str] = (),
        privacy_limit: Optional[st.PrivacyLimit] = None,
        salt: Optional[int] = None,
    ) -> Optional[st.DataSpec]:
        return (
            self.manager()
            .dataspec_rewriter()
            .variant(self, kind, public_context, privacy_limit, salt)
        )

    def variants(self) -> Collection[st.DataSpec]:
        return self.manager().dataspec_rewriter().variants(self)

    def private_queries(self) -> List[st.PrivateQuery]:
        """Return the list of PrivateQueries used in a Dataspec's transform.

        It represents the privacy loss associated with the current computation.

        It can be used by Sarus when a user (Access object) reads a DP dataspec
        to update its accountant. Note that Private Query objects are generated
        with a random uuid so that even if they are submitted multiple times to
        an account, they are only accounted once (ask @cgastaud for more on
        accounting)."""
        return self.manager().dataspec_validator().private_queries(self)

    def value(self) -> st.DataSpecValue:
        return self.manager().value(self)

    async def async_value(self) -> st.DataSpecValue:
        return await self.manager().async_value(self)

    # A Visitor acceptor
    def accept(self, visitor: st.Visitor) -> None:
        visitor.all(self)
        if self.is_transformed():
            visitor.transformed(
                self,
                cast(
                    Transform,
                    self.storage().referrable(
                        self._protobuf.spec.transformed.transform
                    ),
                ),
                *(
                    cast(Scalar, self.storage().referrable(arg))
                    for arg in self._protobuf.spec.transformed.arguments
                ),
                **{
                    name: cast(Scalar, self.storage().referrable(arg))
                    for name, arg in self._protobuf.spec.transformed.named_arguments.items()  # noqa: E501
                },
            )
        else:
            visitor.other(self)

    def dot(self) -> str:
        """return a graphviz representation of the scalar"""

        class Dot(st.Visitor):
            visited: Set[st.DataSpec] = set()
            nodes: Dict[str, Tuple[str, str]] = {}
            edges: Dict[Tuple[str, str], str] = {}

            def transformed(
                self,
                visited: st.DataSpec,
                transform: st.Transform,
                *arguments: st.DataSpec,
                **named_arguments: st.DataSpec,
            ) -> None:
                if visited not in self.visited:
                    if visited.prototype() == sp.Dataset:
                        self.nodes[visited.uuid()] = (
                            visited.name(),
                            "Dataset",
                        )
                    else:
                        self.nodes[visited.uuid()] = (visited.name(), "Scalar")

                    if not visited.is_remote():
                        for argument in arguments:
                            self.edges[
                                (argument.uuid(), visited.uuid())
                            ] = transform.name()
                            argument.accept(self)
                        for _, argument in named_arguments.items():
                            self.edges[
                                (argument.uuid(), visited.uuid())
                            ] = transform.name()
                            argument.accept(self)
                    self.visited.add(visited)

            def other(self, visited: st.DataSpec) -> None:
                if visited.prototype() == sp.Dataset:
                    self.nodes[visited.uuid()] = (
                        visited.name(),
                        "Dataset",
                    )
                else:
                    self.nodes[visited.uuid()] = (visited.name(), "Scalar")

        visitor = Dot()
        self.accept(visitor)
        result = 'digraph {'
        for uuid, (label, node_type) in visitor.nodes.items():
            shape = "polygon" if node_type == "Scalar" else "ellipse"
            result += (
                f'\n"{uuid}" [label="{label} ({uuid[:2]})", shape={shape}];'
            )
        for (u1, u2), label in visitor.edges.items():
            result += f'\n"{u1}" -> "{u2}" [label="{label} ({uuid[:2]})"];'
        result += '}'
        return result

    def attribute(self, name: str) -> Optional[st.Attribute]:
        return self.manager().attribute(name=name, dataspec=self)

    def attributes(self, name: str) -> List[st.Attribute]:
        return self.manager().attributes(name=name, dataspec=self)


def privacy_budget(privacy_limit: st.PrivacyLimit) -> Scalar:
    delta_epsilon_dict = privacy_limit.delta_epsilon_dict()
    return Scalar(
        sp.Scalar(
            name='privacy_budget',
            spec=sp.Scalar.Spec(
                privacy_params=sp.Scalar.PrivacyParameters(
                    points=[
                        sp.Scalar.PrivacyParameters.Point(
                            epsilon=epsilon, delta=delta
                        )
                        for delta, epsilon in delta_epsilon_dict.items()
                    ]
                )
            ),
        )
    )


def random_seed(value: int) -> Scalar:
    return Scalar(
        sp.Scalar(
            name='seed',
            spec=sp.Scalar.Spec(random_seed=sp.Scalar.RandomSeed(value=value)),
        )
    )


def synthetic_model() -> Scalar:
    return Scalar(
        sp.Scalar(
            name='synthetic_model',
            spec=sp.Scalar.Spec(synthetic_model=sp.Scalar.SyntheticModel()),
            properties={'creation_time': str(datetime.datetime.now())},
        )
    )


class Visitor:
    """A visitor class for Scalar"""

    def all(self, visited: Scalar) -> None:
        return

    def transformed(
        self,
        visited: Scalar,
        transform: Transform,
        *arguments: Scalar,
        **named_arguments: Scalar,
    ) -> None:
        return

    def other(self, visited: Scalar) -> None:
        return



from __future__ import annotations

import asyncio
import logging
import traceback
import typing as t

from sarus_data_spec.manager.base import Base
from sarus_data_spec.manager.typing import Computation
from sarus_data_spec.status import status_error_policy
import sarus_data_spec.status as stt
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)
T = t.TypeVar("T")


class BaseComputation(Computation[T]):
    """General class that implements some
    methods of the protocol shared by all task
    computations"""

    task_name = ''

    def __init__(self, computing_manager: Base):
        self._computing_manager = computing_manager

    def computing_manager(self) -> Base:
        return self._computing_manager

    def status(self, dataspec: st.DataSpec) -> t.Optional[st.Status]:
        return self.computing_manager().status(dataspec, self.task_name)

    def launch_task(self, dataspec: st.DataSpec) -> t.Optional[t.Awaitable]:
        """Launch the task computation.

        Returns an optional awaitable that can be used in async functions to
        wait for the task to complete. This can be useful if some managers have
        a more efficient way than statuses to await for the result.
        """
        raise NotImplementedError

    async def task_result(self, dataspec: st.DataSpec, **kwargs: t.Any) -> T:
        """Return the task result.

        This is the main entry point from outide the computation. The call to
        `complete_task` will launch the task if it does not exist and wait for
        it to be ready.

        Here, we assert that the task is ready and then get the result in a
        try/catch block.
        """
        status = await self.complete_task(dataspec=dataspec)
        stage = status.task(self.task_name)
        assert stage
        assert stage.ready()
        try:
            return await self.result_from_stage_properties(
                dataspec, stage.properties(), **kwargs
            )
        except stt.DataSpecErrorStatus as exception:
            stt.error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    'relaunch': str(exception.relaunch),
                },
            )
            raise
        except Exception:
            stt.error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    'relaunch': str(False),
                },
            )
            raise stt.DataSpecErrorStatus((False, traceback.format_exc()))

    async def result_from_stage_properties(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> T:
        """Return the task result by reading cache or computing the value."""
        raise NotImplementedError

    async def complete_task(self, dataspec: st.DataSpec) -> st.Status:
        """Poll the last status for the given task and if no status
        is available either performs the computation or delegates it
        to another manager. Then keeps polling until either the task
        is completed or an error occurs."""

        manager_status = self.status(dataspec)

        if manager_status is None:
            task = self.launch_task(dataspec=dataspec)
            if task is not None:
                # NB: an exception raised in an asyncio task will be reraised
                # in the awaiting code
                await task
            # In the other cases, the complete_task will be reentered with
            # the pending status, resulting in a polling process
            return await self.complete_task(dataspec)

        else:
            last_task = t.cast(st.Stage, manager_status.task(self.task_name))
            if last_task.ready():
                return manager_status
            elif last_task.pending():
                return await self.pending(dataspec)
            elif last_task.processing():
                return await self.processing(dataspec)
            elif last_task.error():
                return await self.error(dataspec)
            else:
                raise ValueError(f"Inconsistent status {manager_status}")

    async def pending(self, dataspec: st.DataSpec) -> st.Status:
        """The behaviour depends on the manager"""
        raise NotImplementedError

    async def processing(self, dataspec: st.DataSpec) -> st.Status:
        """The behaviour depends on the manager"""
        raise NotImplementedError

    async def error(
        self,
        dataspec: st.DataSpec,
    ) -> st.Status:
        """The DataSpec already has an Error status.
        In this case, we clear the statuses so that the
        task can be relaunched in the future.
        """
        status = self.status(dataspec)
        if status is not None:
            stage = status.task(self.task_name)
            assert stage
            should_clear = status_error_policy(stage=stage)
            if should_clear:
                status.clear_task(self.task_name)
                return await self.complete_task(dataspec=dataspec)
            raise stt.DataSpecErrorStatus(
                (
                    stage.properties()["relaunch"] == str(True),
                    stage.properties()["message"],
                )
            )
        return await self.complete_task(dataspec=dataspec)

    async def wait_for_computation(
        self,
        dataspec: st.DataSpec,
        current_stage: str,
        timeout: int = 300,
        max_delay: int = 10,
    ) -> st.Stage:
        """Utility to wait for the availability of a computation, by polling
        its status for at most `timeout` seconds, with a period of at most
        `max_delay` seconds between each attempt. Delay grows exponentially
        at first."""
        delay = min(1, timeout, max_delay)
        total_wait = 0
        while total_wait < timeout:
            status = self.status(dataspec=dataspec)
            assert status
            stage = status.task(self.task_name)
            assert stage
            if stage.stage() == current_stage:
                logger.debug(f'POLLING {self.task_name} {dataspec.uuid()}')
                await asyncio.sleep(delay)
                total_wait += delay
                delay = min(2 * delay, max_delay, timeout - total_wait)
            else:
                break
        assert stage
        return stage

    def force_launch(self, dataspec: st.DataSpec) -> None:
        """Method to force launch when an existing
        status with relaunch=true exists. it creates
        a new status without the task in question
        and calls launch task"""
        status = self.status(dataspec=dataspec)
        if status is not None:
            stage = status.task(self.task_name)
            assert stage
            if stage.error() and status_error_policy(stage=stage):
                status.clear_task(self.task_name)
        self.launch_task(dataspec=dataspec)


class ErrorCatchingAsyncIterator:
    """Wrap an AsyncIterator and catches potential errors.

    When an error occurs, this sets the Dataspec status to error
    accordingly.
    """

    def __init__(
        self,
        ait: t.AsyncIterator,
        dataspec: st.DataSpec,
        computation: BaseComputation,
    ):
        self.ait = ait
        self.computation = computation
        self.dataspec = dataspec
        self.agen: t.Optional[t.AsyncIterator] = None

    def __aiter__(self) -> t.AsyncIterator:
        return self.ait

    async def __anext__(self) -> t.Any:
        try:
            batch = await self.ait.__anext__()
        except StopAsyncIteration:
            raise
        except stt.DataSpecErrorStatus as exception:
            stt.error(
                dataspec=self.dataspec,
                manager=self.computation.computing_manager(),
                task=self.computation.task_name,
                properties={
                    "message": traceback.format_exc(),
                    'relaunch': str(exception.relaunch),
                },
            )
            raise
        except Exception:
            stt.error(
                dataspec=self.dataspec,
                manager=self.computation.computing_manager(),
                task=self.computation.task_name,
                properties={
                    "message": traceback.format_exc(),
                    'relaunch': str(False),
                },
            )
            raise stt.DataSpecErrorStatus((False, traceback.format_exc()))

        else:
            return batch



from __future__ import annotations

from typing import Collection, List, Optional, cast
import json
import logging

from sarus_data_spec.attribute import attach_properties
from sarus_data_spec.constants import (
    IS_VALID,
    NO_TOKEN,
    PEP_TOKEN,
    PRIVATE_QUERY,
    PUBLIC,
)
from sarus_data_spec.dataspec_validator.privacy_limit import DeltaEpsilonLimit
from sarus_data_spec.manager.async_utils import sync
from sarus_data_spec.manager.ops.processor import routing
from sarus_data_spec.manager.ops.processor.external.external_op import (
    external_implementation,
)
from sarus_data_spec.protobuf.utilities import dejson
from sarus_data_spec.protobuf.utilities import json as proto_to_json
from sarus_data_spec.storage.typing import Storage
from sarus_data_spec.variant_constraint import (
    pep_constraint,
    public_constraint,
    syn_constraint,
)
import sarus_data_spec.dataspec_validator.simple_rules as verification_rules
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

try:
    from sarus_differential_privacy.protobuf.private_query_pb2 import (
        PrivateQuery as ProtoPrivateQuery,
    )
    from sarus_differential_privacy.query import BasePrivateQuery

except ImportError:
    # Warning raised in typing.py
    pass

logger = logging.getLogger(__name__)


class BaseDataspecValidator:
    def __init__(self, storage: Storage):
        self._storage = storage

    def storage(self) -> Storage:
        return self._storage

    def verified_constraints(
        self, dataspec: st.DataSpec
    ) -> List[st.VariantConstraint]:
        """Return the list of VariantConstraints attached to a DataSpec.

        A VariantConstraint attached to a DataSpec means that the DataSpec
        verifies the constraint.
        """
        constraints = self.storage().referring(
            dataspec, type_name=sp.type_name(sp.VariantConstraint)
        )
        return cast(List[st.VariantConstraint], list(constraints))

    def verifies(
        self,
        variant_constraint: st.VariantConstraint,
        kind: st.ConstraintKind,
        public_context: Collection[str],
        privacy_limit: Optional[st.PrivacyLimit],
        salt: Optional[int] = None,
    ) -> Optional[bool]:
        """Check if the constraint attached to a Dataspec meets requirements.

        This function is useful because comparisons are not straightforwards.
        For instance, a Dataspec might have the variant constraint SYNTHETIC
        attached to it. This synthetic dataspec also verifies the DP constraint
        and the PUBLIC constraint.

        Args:
            variant_constraint: VariantConstraint attached to the Dataspec
            kind: constraint kind to verify compliance with
            public_context: actual current public context
            epsilon: current privacy consumed
        """
        return verification_rules.verifies(
            variant_constraint=variant_constraint,
            kind=kind,
            public_context=public_context,
            privacy_limit=privacy_limit,
            salt=salt,
        )

    def is_dp(self, dataspec: st.DataSpec) -> bool:
        """Return True if the dataspec is the result of a DP transform.

        This is a simple implementation. This function checks if the
        dataspec's transform has a privacy budget and a random seed as an
        argument.
        """
        if not dataspec.is_transformed():
            return False

        parents, kwparents = dataspec.parents()
        parents = list(parents) + list(kwparents.values())
        scalars = [
            cast(st.Scalar, parent)
            for parent in parents
            if parent.prototype() == sp.Scalar
        ]
        has_budget = (
            len([scalar for scalar in scalars if scalar.is_privacy_params()])
            == 1
        )
        has_seed = (
            len([scalar for scalar in scalars if scalar.is_random_seed()]) == 1
        )
        return has_budget and has_seed

    def is_synthetic(self, dataspec: st.DataSpec) -> bool:
        """Return True if the dataspec is synthetic.

        This functions creates a VariantConstraint on the DataSpec to cache
        the SYNTHETIC constraint.
        """
        # TODO fetch real context and epsilon
        public_context: Collection[str] = []
        privacy_limit = None
        kind = st.ConstraintKind.SYNTHETIC

        self.is_public(dataspec)

        for constraint in self.verified_constraints(dataspec):
            check_constraint = self.verifies(
                constraint, kind, public_context, privacy_limit
            )
            if check_constraint is not None:
                return check_constraint

        # Determine is the Dataspec is synthetic
        if dataspec.is_transformed():
            transform = dataspec.transform()
            if transform.protobuf().spec.HasField("synthetic"):
                is_synthetic = True
            else:
                # Returns true if the DataSpec derives only from synthetic
                args_parents, kwargs_parents = dataspec.parents()
                is_synthetic = all(
                    [
                        self.is_synthetic(ds)
                        for ds in args_parents
                        if isinstance(ds, st.DataSpec)
                    ]
                    + [
                        self.is_synthetic(ds)
                        for ds in kwargs_parents.values()
                        if isinstance(ds, st.DataSpec)
                    ]
                )
        else:
            is_synthetic = False

        # save variant constraint
        syn_constraint(dataspec, is_synthetic=is_synthetic)

        return is_synthetic

    def is_public(self, dataspec: st.DataSpec) -> bool:
        """Return True if the dataspec is public.

        Some DataSpecs are intrinsically Public, this is the case if they are
        freely available externally, they can be tagged so and will never be
        considered otherwise.

        This function returns True in the following cases:
        - The dataspec is an ML model
        - The dataspec is transformed but all its inputs are public

        This functions creates a VariantConstraint on the DataSpec to cache the
        PUBLIC constraint.
        """
        # TODO fetch real context and epsilon
        public_context: Collection[str] = []
        privacy_limit = DeltaEpsilonLimit({0.0: 0.0})
        kind = st.ConstraintKind.PUBLIC

        for constraint in self.verified_constraints(dataspec):
            check_constraint = self.verifies(
                constraint, kind, public_context, privacy_limit
            )
            if check_constraint is not None:
                return check_constraint

        # Determine is the Dataspec is public
        if dataspec.is_transformed():
            # Returns true if the DataSpec derives only from public
            if (
                dataspec.transform().is_external()
                or dataspec.prototype() == sp.Scalar
            ):
                args_parents, kwargs_parents = dataspec.parents()
                is_public = all(
                    [
                        self.is_public(ds)
                        for ds in args_parents
                        if isinstance(ds, st.DataSpec)
                    ]
                    + [
                        self.is_public(ds)
                        for ds in kwargs_parents.values()
                        if isinstance(ds, st.DataSpec)
                    ]
                )
            else:
                assert dataspec.prototype() == sp.Dataset
                # For a standard transform, all tables must be
                # public in the schema to have a public dataset
                dataset = cast(st.Dataset, dataspec)
                schema = dataset.schema()
                is_public = schema.data_type().properties()[PUBLIC] == str(
                    True
                )

        elif dataspec.prototype() == sp.Scalar:
            scalar = cast(st.Scalar, dataspec)
            assert (
                scalar.is_random_seed()
                or scalar.is_privacy_params()
                or scalar.is_synthetic_model()
            )
            is_public = True
        else:
            is_public = False

        # save variant constraint
        public_constraint(dataspec, is_public)

        return is_public

    def pep_token(self, dataspec: st.DataSpec) -> Optional[str]:
        """Return a token if the dataspec is PEP, otherwise return None.

        DataSpec.pep_token() returns a PEP token if the dataset is PEP and None
        otherwise. The PEP token is stored in the properties of the
        VariantConstraint. It is a hash initialized with a value when the
        Dataset is protected.

        If a transform does not preserve the PEID then the token is set to None
        If a transform preserves the PEID assignment but changes the rows (e.g.
        sample, shuffle, filter,...) then the token's value is changed If a
        transform does not change the rows (e.g. selecting a column, adding a
        scalar,...) then the token is passed without change

        A Dataspec is PEP if its PEP token is not None. Two PEP Dataspecs are
        aligned (i.e. they have the same number of rows and all their rows have
        the same PEID) if their tokens are equal.
        """
        if dataspec.prototype() == sp.Scalar:
            return None

        dataset = cast(st.Dataset, dataspec)

        # TODO fetch real context and budget
        public_context: Collection[str] = []
        privacy_limit = DeltaEpsilonLimit({0.0: 0.0})
        kind = st.ConstraintKind.PEP

        for constraint in self.verified_constraints(dataspec):
            check_constraint = self.verifies(
                constraint, kind, public_context, privacy_limit
            )
            if check_constraint is not None:
                if check_constraint:
                    return constraint.properties()[PEP_TOKEN]
                else:
                    return None

        # Compute the PEP token
        if not dataset.is_transformed():
            return None

        transform = dataset.transform()
        _, StaticChecker = routing.get_dataset_op(transform)
        pep_token = StaticChecker(dataset).pep_token(public_context)
        if pep_token is None:
            pep_token = NO_TOKEN

        pep_constraint(
            dataspec=dataset,
            token=pep_token,
            required_context=[],
            privacy_limit=privacy_limit,
        )

        return None if pep_token == NO_TOKEN else pep_token

    def private_queries(self, dataspec: st.DataSpec) -> List[st.PrivateQuery]:
        """Return the list of PrivateQueries used in a Dataspec's transform.

        It represents the privacy loss associated with the current computation.

        It can be used by Sarus when a user (Access object) reads a DP dataspec
        to update its accountant. Note that Private Query objects are generated
        with a random uuid so that even if they are submitted multiple times to
        an account, they are only accounted once (ask @cgastaud for more on
        accounting).
        """
        attribute = dataspec.attribute(name=PRIVATE_QUERY)
        # Already computed
        if attribute is not None:
            private_query_str = attribute[PRIVATE_QUERY]
            protos = [
                cast(ProtoPrivateQuery, dejson(q))
                for q in json.loads(private_query_str)
            ]
            return cast(
                List[st.PrivateQuery],
                BasePrivateQuery.from_protobuf(protos),
            )

        # Compute private queries
        if not dataspec.is_transformed():
            private_queries = []
        else:
            if dataspec.prototype() == sp.Dataset:
                dataset = cast(st.Dataset, dataspec)
                private_queries = sync(
                    routing.TransformedDataset(dataset).private_queries()
                )
            else:
                scalar = cast(st.Scalar, dataspec)
                private_queries = sync(
                    routing.TransformedScalar(scalar).private_queries()
                )

        # Cache in an attribute
        subqueries = [
            proto_to_json(query.protobuf()) for query in private_queries
        ]
        attach_properties(
            dataspec,
            properties={PRIVATE_QUERY: json.dumps(subqueries)},
            name=PRIVATE_QUERY,
        )

        return private_queries

    def has_valid_transform(self, dataspec: st.DataSpec) -> bool:
        """Check that the transform of a dataspec is valid with the input
        parameters.
        """
        if not dataspec.is_transformed():
            return True

        transform = dataspec.transform()
        # TODO: remove condition is_external when standard
        # transforms has signature
        if transform.is_external():
            implementation = external_implementation(transform)
            try:
                bound_signature = implementation.signature().bind_dataspec(
                    dataspec
                )
                bound_signature.static_validation()
                return True
            except (ValueError, TypeError):
                return False
        else:
            return True

    def is_valid(self, dataspec: st.DataSpec) -> bool:
        """
        Check that the dataspec is validating certain conditions: valid
        transforms, valid parents, valid sources.

        This function creates an attributes on the DataSpec to cache the
        validity of the dataspecs.
        The source dataspec are validated during the onboarding process.
        """
        # Valid attribute
        is_valid_att = dataspec.attribute(IS_VALID)
        if is_valid_att is not None:
            return is_valid_att[IS_VALID] == str(True)

        # The unvalidated source is not valid.
        if not dataspec.is_transformed():
            if dataspec.is_public():
                is_valid = True
            else:
                is_valid = False
        else:
            # Valid transform
            if self.has_valid_transform(dataspec):
                # Valid parents:
                parents, kwparents = dataspec.parents()
                parents = list(parents) + list(kwparents.values())
                is_valid = all(
                    self.is_valid(parent)
                    for parent in parents
                    if isinstance(parent, st.DataSpec)
                )
            else:
                is_valid = False

        # Only one non-public source max.
        sources_ds = dataspec.sources(sp.type_name(sp.Dataset))
        admin_sources = [
            source for source in sources_ds if not source.is_public()
        ]
        if len(admin_sources) > 1:
            is_valid = False

        attach_properties(
            dataspec,
            name=IS_VALID,
            properties={IS_VALID: str(is_valid)},
        )
        return is_valid

    def is_dp_writable(self, dataspec: st.DataSpec) -> bool:
        raise NotImplementedError

    def is_pep_writable(self, dataspec: st.DataSpec) -> bool:
        raise NotImplementedError

    def is_publishable(self, dataspec: st.DataSpec) -> bool:
        raise NotImplementedError

    def is_published(self, dataspec: st.DataSpec) -> bool:
        raise NotImplementedError

    def rewritten_pep_token(self, dataspec: st.DataSpec) -> Optional[str]:
        raise NotImplementedError



from typing import Collection, List, Optional, Tuple, cast
import logging

from sarus_data_spec.constants import (
    IS_PUBLIC,
    IS_SYNTHETIC,
    NO_TOKEN,
    PEP_TOKEN,
    SALT,
)
import sarus_data_spec.typing as st

ArgStruct = Tuple[List[int], List[str]]
logger = logging.getLogger(__name__)


def verifies(
    variant_constraint: st.VariantConstraint,
    kind: st.ConstraintKind,
    public_context: Collection[str],
    privacy_limit: Optional[st.PrivacyLimit],
    salt: Optional[int] = None,
) -> Optional[bool]:
    if kind == st.ConstraintKind.PUBLIC:
        return verifies_public(variant_constraint=variant_constraint)

    elif kind == st.ConstraintKind.SYNTHETIC:
        return verifies_synthetic(variant_constraint=variant_constraint)

    elif kind == st.ConstraintKind.MOCK:
        return verifies_mock(variant_constraint=variant_constraint)

    elif kind == st.ConstraintKind.DP:
        return verifies_dp(
            variant_constraint=variant_constraint,
            privacy_limit=privacy_limit,
            salt=salt,
        )
    elif kind == st.ConstraintKind.PEP:
        return verifies_pep(variant_constraint=variant_constraint)
    else:  # kind == st.ConstraintKind.PEP_FOR_REWRITING:
        return verifies_pep_for_rewriting(
            variant_constraint=variant_constraint
        )


def verifies_public(
    variant_constraint: st.VariantConstraint,
) -> Optional[bool]:
    kind = variant_constraint.constraint_kind()
    if kind == st.ConstraintKind.PUBLIC:
        return variant_constraint.properties()[IS_PUBLIC] == str(True)
    else:
        return None


def verifies_synthetic(
    variant_constraint: st.VariantConstraint,
) -> Optional[bool]:
    kind = variant_constraint.constraint_kind()
    if kind == st.ConstraintKind.SYNTHETIC:
        return variant_constraint.properties()[IS_SYNTHETIC] == str(True)
    elif kind == st.ConstraintKind.PUBLIC:
        if variant_constraint.properties()[IS_PUBLIC] == str(True):
            return True
        else:
            return None
    else:
        return None


def verifies_mock(variant_constraint: st.VariantConstraint) -> Optional[bool]:
    kind = variant_constraint.constraint_kind()
    if kind == st.ConstraintKind.MOCK:
        return True
    elif kind == st.ConstraintKind.PUBLIC:
        if variant_constraint.properties()[IS_PUBLIC] == str(True):
            return True
        else:
            return None
    else:
        return None


def verifies_pep(
    variant_constraint: st.VariantConstraint,
) -> Optional[bool]:
    """If we attached a PEP constraint to a dataspec then it is PEP.

    NB: for now we don't check the context nor the privacy limit
    """
    kind = variant_constraint.constraint_kind()
    if kind == st.ConstraintKind.PEP:
        stored_token = variant_constraint.properties()[PEP_TOKEN]
        return False if stored_token == NO_TOKEN else True
    else:
        return None


def verifies_pep_for_rewriting(
    variant_constraint: st.VariantConstraint,
) -> Optional[bool]:
    kind = variant_constraint.constraint_kind()
    if kind == st.ConstraintKind.PEP_FOR_REWRITING:
        stored_token = variant_constraint.properties()[PEP_TOKEN]
        return False if stored_token == NO_TOKEN else True
    else:
        return None


def verifies_dp(
    variant_constraint: st.VariantConstraint,
    privacy_limit: Optional[st.PrivacyLimit],
    salt: Optional[int] = None,
) -> Optional[bool]:
    """Check if a variant constraint satisfies a DP profile.

    For now, return True only for strict equality.
    """
    if privacy_limit is None:
        raise ValueError(
            "Input privacy limit required when checking against DP."
        )

    kind = variant_constraint.constraint_kind()
    if kind != st.ConstraintKind.DP:
        return None

    constraint_privacy_limit = variant_constraint.privacy_limit()
    if constraint_privacy_limit is None:
        raise ValueError(
            "Found a DP constraint without a privacy limit "
            "when checking against DP."
        )

    constraint_salt = variant_constraint.properties().get(SALT)
    if salt and str(salt) != constraint_salt:
        return False

    return cast(
        bool,
        privacy_limit.delta_epsilon_dict()
        == constraint_privacy_limit.delta_epsilon_dict(),
    )



import typing as t
import warnings

import pyarrow as pa

from sarus_data_spec.attribute import attach_properties
from sarus_data_spec.bounds import bounds as bounds_builder
from sarus_data_spec.marginals import marginals as marginals_builder
from sarus_data_spec.multiplicity import multiplicity as multiplicity_builder
from sarus_data_spec.size import size as size_builder
import sarus_data_spec.typing as st

try:
    from sarus_synthetic_data.synthetic_generator.generator import (
        SyntheticGenerator,
    )

except ModuleNotFoundError:
    warnings.warn(
        'sarus-synthetic-data Module not found, synthetic data operations not '
        'available '
    )

from sarus_data_spec.constants import (
    DATASET_SLUGNAME,
    SYNTHETIC_MODEL,
    SYNTHETIC_TASK,
    TRAIN_CORRELATIONS,
)
from sarus_data_spec.dataset import Dataset

try:
    from sarus_data_spec.manager.ops.source.query_builder import (
        synthetic_parameters,
    )
except ModuleNotFoundError:
    warnings.warn(
        "synthetic_parameters not found, "
        "synthetic data operations not available "
    )
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.schema import schema

from .standard_op import (
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
    StandardScalarImplementation,
    StandardScalarStaticChecker,
)

MAX_SIZE = 1e6  # TODO: in sarus_data_spec.constants ?


def convert_array_to_table(
    schema_type: st.Type, arrow_data: pa.array
) -> pa.Array:
    """Given a PyArrow array, returns a correctly-defined Table."""

    class ArrayToTable(st.TypeVisitor):
        """Handles both configuration: a dataset as a Struct or as an Union."""

        result = None

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            names = list(fields.keys())
            self.result = pa.Table.from_arrays(
                arrays=arrow_data.flatten(), names=names
            )

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            names = list(fields.keys())
            arrs = arrow_data.flatten()
            schema = pa.schema(
                [
                    pa.field(name, type=arr.type)
                    for arr, name in zip(arrs[:-1], names)
                ]
            )
            schema = schema.append(
                pa.field('field_selected', type=pa.string(), nullable=False)
            )
            self.result = pa.Table.from_arrays(
                arrays=arrow_data.flatten(), schema=schema
            )

        def Null(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            raise NotImplementedError

        def Unit(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            raise NotImplementedError

        def Boolean(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            raise NotImplementedError

        def Id(
            self,
            unique: bool,
            reference: t.Optional[st.Path] = None,
            base: t.Optional[st.IdBase] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Integer(
            self,
            min: int,
            max: int,
            base: st.IntegerBase,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Enum(
            self,
            name: str,
            name_values: t.Sequence[t.Tuple[str, int]],
            ordered: bool,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Float(
            self,
            min: float,
            max: float,
            base: st.FloatBase,
            possible_values: t.Iterable[float],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Text(
            self,
            encoding: str,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Bytes(
            self, properties: t.Optional[t.Mapping[str, str]] = None
        ) -> None:
            raise NotImplementedError

        def Optional(
            self,
            type: st.Type,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Datetime(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DatetimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Time(
            self,
            format: str,
            min: str,
            max: str,
            base: st.TimeBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Date(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DateBase,
            possible_values: t.Iterable[str],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Duration(
            self,
            unit: str,
            min: int,
            max: int,
            possible_values: t.Iterable[int],
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Constrained(
            self,
            type: st.Type,
            constraint: st.Predicate,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Hypothesis(
            self,
            *types: t.Tuple[st.Type, float],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

    visitor = ArrayToTable()
    schema_type.accept(visitor)
    return visitor.result


async def async_iter_arrow(
    iterator: t.Iterator[pa.RecordBatch],
) -> t.AsyncIterator[pa.RecordBatch]:
    """Async generator from the synthetic data iterator."""
    for batch in iterator:
        yield batch
    return


class SyntheticStaticChecker(StandardDatasetStaticChecker):
    def pep_token(self, public_context: t.Collection[str]) -> t.Optional[str]:
        # TODO add pep token when the synthetic data is actually protected
        return None

    def rewritten_pep_token(
        self, public_context: t.Collection[str]
    ) -> t.Optional[str]:
        # TODO add pep token when the synthetic data is actually protected
        return None

    async def schema(self) -> st.Schema:
        parent_schema = await self.parent_schema()
        return schema(
            self.dataset,
            schema_type=parent_schema.data_type(),
            properties=parent_schema.properties(),
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
        )


class Synthetic(StandardDatasetImplementation):
    """Create a Synthetic op class for is_pep."""

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        dataset = self.dataset
        parents, parents_dict = dataset.parents()

        # Forcing the marginals to be computed first
        parent = t.cast(Dataset, parents[0])
        _ = await parent.manager().async_marginals(parent)

        # Budget
        budget_param = parents_dict['sd_budget']
        budget = t.cast(
            t.Tuple[float, float],
            await dataset.manager().async_value(t.cast(Scalar, budget_param)),
        )

        # Model
        correlations_scalar = t.cast(Scalar, parents_dict['synthetic_model'])
        train_correlations = t.cast(
            bool, await dataset.manager().async_value(correlations_scalar)
        )

        # Generator params
        generator_params = await synthetic_parameters(
            dataset,
            sd_budget=budget,
            task=SYNTHETIC_TASK,
            train_correlations=train_correlations,
        )
        # Links computation
        _ = await self.dataset.manager().async_links(self.dataset)
        # compute
        generator = SyntheticGenerator(dataset, generator_params.generator)
        dataset_schema = await dataset.manager().async_schema(dataset)
        datatype = dataset_schema.type()
        generator.train()
        sample = generator.sample()
        table = convert_array_to_table(datatype, sample)
        return async_iter_arrow(table.to_batches(max_chunksize=batch_size))

    async def size(self) -> st.Size:
        parent_size = await self.parent_size()
        return size_builder(self.dataset, parent_size.statistics())

    async def multiplicity(self) -> st.Multiplicity:
        parent_multiplicity = await self.parent_multiplicity()
        return multiplicity_builder(
            self.dataset, parent_multiplicity.statistics()
        )

    async def bounds(self) -> st.Bounds:
        parent_bounds = await self.parent_bounds()
        return bounds_builder(self.dataset, parent_bounds.statistics())

    async def marginals(self) -> st.Marginals:
        parent_marginals = await self.parent_marginals()
        return marginals_builder(self.dataset, parent_marginals.statistics())


class SamplingRatiosStaticChecker(StandardScalarStaticChecker):
    ...


class SamplingRatios(StandardScalarImplementation):
    """Computes the sampling ratios for the SD
    of the dataspec given the total budget"""

    async def value(self) -> t.Any:
        dataset = t.cast(st.Dataset, self.parent())
        out = {}
        for table_path in (await self.parent_schema()).tables():
            sizes = await dataset.manager().async_size(dataset)
            assert sizes
            stat = sizes.statistics().nodes_statistics(table_path)[0]
            out[table_path] = min(1, MAX_SIZE / stat.size())
        return out


class SyntheticModelStaticChecker(StandardScalarStaticChecker):
    ...


class SyntheticModel(StandardScalarImplementation):
    """Computes the synthetic model to use"""

    async def value(self) -> t.Any:
        attribute = self.scalar.attribute(name=SYNTHETIC_MODEL)
        if attribute is None:
            attach_properties(
                self.scalar,
                name=SYNTHETIC_MODEL,
                properties={TRAIN_CORRELATIONS: str(True)},
            )
            return True
        return attribute.properties()[TRAIN_CORRELATIONS] == str(True)



from __future__ import annotations

from typing import (
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Protocol,
    Tuple,
    runtime_checkable,
)
import typing as t

import pandas as pd
import pyarrow as pa

try:
    import tensorflow as tf
except ModuleNotFoundError:
    pass  # Warning is displayed by typing.py

import warnings

from sarus_data_spec.storage.typing import HasStorage
import sarus_data_spec.dataspec_rewriter.typing as sdrt
import sarus_data_spec.dataspec_validator.typing as sdvt
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

try:
    import sqlalchemy as sa

    sa_engine = sa.engine.Engine
except ModuleNotFoundError:
    warnings.warn("Sqlalchemy not installed, cannot send sql queries")
    sa_engine = t.Any  # type: ignore


@runtime_checkable
class Manager(st.Referrable[sp.Manager], HasStorage, Protocol):
    """Provide the dataset functionalities"""

    def to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> t.Iterator[pa.RecordBatch]:
        """Synchronous method based on async_to_arrow
        that returns an iterator of arrow batches
        for the input dataset"""
        ...

    async def async_to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> AsyncIterator[pa.RecordBatch]:
        """Asynchronous method. It orchestrates how
        the iterator is obtained: it can either be delegated
        via arrow_task and the result polled, or computed directly
        it via the op"""
        ...

    def schema(self, dataset: st.Dataset) -> st.Schema:
        """Synchronous method that returns the schema of a
        dataspec. Based on the asynchronous version"""
        ...

    async def async_schema(self, dataset: st.Dataset) -> st.Schema:
        """Asynchronous method that returns the schema of a
        dataspec. The computation can be either delegated to
        another manager via schema_task and the result polled
        or executed directly via async_schema_ops"""
        ...

    def value(self, scalar: st.Scalar) -> st.DataSpecValue:
        """Synchronous method that returns the value of a
        scalar. Based on the asynchronous version"""
        ...

    async def async_value(self, scalar: st.Scalar) -> st.DataSpecValue:
        """Asynchronous method that returns the value of a
        scalar. The computation can be either delegated to
        another manager via value_task and the result polled
        or executed directly via async_value_ops"""
        ...

    def prepare(self, dataspec: st.DataSpec) -> None:
        """Make sure a Dataspec is ready."""
        ...

    async def async_prepare(self, dataspec: st.DataSpec) -> None:
        """Make sure a Dataspec is ready asynchronously."""
        ...

    async def async_prepare_parents(self, dataspec: st.DataSpec) -> None:
        """Prepare all the parents of a Dataspec."""
        ...

    def sql_prepare(self, dataset: st.Dataset) -> None:
        """Make sure a dataset is sql ready"""
        ...

    async def async_sql_prepare(self, dataset: st.Dataset) -> None:
        """Make sure a dataset is sql ready asynchronously."""
        ...

    async def async_sql_prepare_parents(self, dataset: st.Dataset) -> None:
        """SQL prepare all the parents of a dataset. It should sql_prepare
        dataset parents and prepare Scalars parents.
        """
        ...

    def cache_scalar(self, scalar: st.Scalar) -> None:
        """Synchronous scalar caching"""
        ...

    async def async_cache_scalar(self, scalar: st.Scalar) -> None:
        """Asynchronous scalar caching"""
        ...

    def to_parquet(self, dataset: st.Dataset) -> None:
        """Synchronous parquet caching"""
        ...

    async def async_to_parquet(self, dataset: st.Dataset) -> None:
        """Asynchronous parquet caching"""
        ...

    def parquet_dir(self) -> str:
        ...

    def marginals(self, dataset: st.Dataset) -> st.Marginals:
        ...

    async def async_marginals(self, dataset: st.Dataset) -> st.Marginals:
        ...

    def bounds(self, dataset: st.Dataset) -> st.Bounds:
        ...

    async def async_bounds(self, dataset: st.Dataset) -> st.Bounds:
        ...

    def size(self, dataset: st.Dataset) -> st.Size:
        ...

    async def async_size(self, dataset: st.Dataset) -> st.Size:
        ...

    def multiplicity(self, dataset: st.Dataset) -> st.Multiplicity:
        ...

    async def async_multiplicity(self, dataset: st.Dataset) -> st.Multiplicity:
        ...

    def to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        ...

    async def async_to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        ...

    async def async_to(
        self, dataset: st.Dataset, kind: t.Type, drop_admin: bool = True
    ) -> st.DatasetCastable:
        """Casts a Dataset to a Python type passed as argument."""
        ...

    def to(
        self, dataset: st.Dataset, kind: t.Type, drop_admin: bool = True
    ) -> st.DatasetCastable:
        ...

    def to_tensorflow(self, dataset: st.Dataset) -> tf.data.Dataset:
        ...

    async def async_to_tensorflow(
        self, dataset: st.Dataset
    ) -> tf.data.Dataset:
        ...

    def to_sql(self, dataset: st.Dataset) -> None:
        ...

    async def async_to_sql(self, dataset: st.Dataset) -> None:
        ...

    def status(
        self, dataspec: st.DataSpec, task_name: t.Optional[str] = None
    ) -> t.Optional[st.Status]:
        ...

    def dataspec_rewriter(self) -> sdrt.DataspecRewriter:
        ...

    def dataspec_validator(self) -> sdvt.DataspecValidator:
        ...

    def is_remote(self, dataspec: st.DataSpec) -> bool:
        """Is the dataspec a remotely defined dataset."""
        ...

    def infer_output_type(
        self,
        transform: st.Transform,
        *arguments: t.Union[st.DataSpec, st.Transform],
        **named_arguments: t.Union[st.DataSpec, st.Transform],
    ) -> Tuple[str, Callable[[st.DataSpec], None]]:
        ...

    def foreign_keys(self, dataset: st.Dataset) -> Dict[st.Path, st.Path]:
        ...

    async def async_foreign_keys(
        self, dataset: st.Dataset
    ) -> Dict[st.Path, st.Path]:
        ...

    async def async_primary_keys(self, dataset: st.Dataset) -> List[st.Path]:
        ...

    def primary_keys(self, dataset: st.Dataset) -> List[st.Path]:
        ...

    def sql(
        self,
        dataset: st.Dataset,
        query: t.Union[str, t.Dict[str, t.Any]],
        dialect: Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
    ) -> Iterator[pa.RecordBatch]:
        ...

    async def async_sql(
        self,
        dataset: st.Dataset,
        query: t.Union[str, t.Dict[str, t.Any]],
        dialect: Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
        result_type: t.Optional[st.Type] = None,
    ) -> AsyncIterator[pa.RecordBatch]:
        ...

    async def execute_sql_query(
        self,
        dataset: st.Dataset,
        caching_properties: t.Mapping[str, str],
        query: t.Union[str, t.Dict[str, t.Any]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
        result_type: t.Optional[st.Type] = None,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        ...

    async def async_sql_op(
        self,
        dataset: st.Dataset,
        query: t.Union[str, t.Dict[str, t.Any]],
        dialect: t.Optional[st.SQLDialect] = None,
        batch_size: int = 10000,
        result_type: t.Optional[st.Type] = None,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        ...

    def is_big_data(self, dataset: st.Dataset) -> bool:
        ...

    def is_cached(self, dataspec: st.DataSpec) -> bool:
        """Returns whether a dataspec should be cached
        or not"""
        ...

    def is_pushed_to_sql(self, dataspec: st.DataSpec) -> bool:
        """Returns whether a dataspec should be pushed to sql
        or not"""
        ...

    def attribute(
        self, name: str, dataspec: st.DataSpec
    ) -> t.Optional[st.Attribute]:
        ...

    def attributes(
        self, name: str, dataspec: st.DataSpec
    ) -> t.List[st.Attribute]:
        ...

    def links(self, dataset: st.Dataset) -> st.Links:
        ...

    async def async_links(self, dataset: st.Dataset) -> st.Links:
        ...

    def sql_pushing_schema_prefix(self, dataset: st.Dataset) -> str:
        ...

    def engine(self, uri: str) -> sa_engine:
        ...

    def mock_value(
        self,
        transform: st.Transform,
        *arguments: st.DataSpec,
        **named_arguments: st.DataSpec,
    ) -> t.Any:
        """Compute the mock value of an external transform applied on
        Dataspecs.
        """

    def composed_callable(
        self, transform: st.Transform
    ) -> t.Callable[..., t.Any]:
        """Return a Python callable of a composed transform."""


@runtime_checkable
class HasManager(Protocol):
    """Has a manager."""

    def manager(self) -> Manager:
        """Return a manager (usually a singleton)."""
        ...


T = t.TypeVar("T", covariant=True)


class Computation(t.Protocol[T]):
    """Protocol for classes that perform tasks computations.
    It sets how computations are scheduled and launched
    depending on statuses. A computation is mainly defined by two methods:
     - launch : a method that does not return a value but
     that only has side effects, changing either the storage or the cache
    and that updates statuses during the process
    - result: a method that allows to get the value of the computation
    either by reading the cache/storage or via the ops.

    Furthermore, a computation has a method to monitor task completion.
    """

    task_name: str = ''

    def launch_task(self, dataspec: st.DataSpec) -> t.Optional[t.Awaitable]:
        """This methods launches a task in the background
        but returns immediately without waiting for the
        result. It updates the statuses during its process."""
        ...

    async def task_result(self, dataspec: st.DataSpec, **kwargs: t.Any) -> T:
        """Returns the value for the given computed task. It either
        retrieves it from the cache or computes it via the ops."""
        ...

    async def complete_task(self, dataspec: st.DataSpec) -> st.Status:
        """Monitors a task: it launches it if there is no status
        and then polls until it is ready/error"""
        ...



from datetime import timedelta
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    List,
    Literal,
    Optional,
    Union,
)

from pandas._libs.tslibs import BaseOffset
from pandas.core.window.indexers import BaseIndexer
import pandas as pd
import pandas._typing as pdt

from sarus_data_spec.dataspec_validator.parameter_kind import DATASPEC
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

GroupBy = Union[
    pd.core.groupby.DataFrameGroupBy,
    pd.core.groupby.SeriesGroupBy,
]


class pd_agg_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_AGG_GROUPBY"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GroupBy,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="func",
            annotation=Optional[Union[Callable, Dict, List]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.agg(**kwargs)


class pd_count_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_COUNT_GROUPBY"
    _dp_equivalent_id = "pandas.PD_COUNT_GROUPBY_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GroupBy,
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.count(**kwargs)


class pd_max_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_MAX_GROUPBY"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GroupBy,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="min_count",
            annotation=int,
            default=-1,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.max(**kwargs)


class pd_mean_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_MEAN_GROUPBY"
    _dp_equivalent_id = "pandas.PD_MEAN_GROUPBY_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GroupBy,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.mean(**kwargs)


class pd_min_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_MIN_GROUPBY"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GroupBy,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="min_count",
            annotation=int,
            default=-1,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.min(**kwargs)


class pd_rolling_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_ROLLING_GROUPBY"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GroupBy,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="window",
            annotation=Union[int, timedelta, BaseOffset, BaseIndexer],
        ),
        SarusParameter(
            name="min_periods",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="center",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="win_type",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="on",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="closed",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="method",
            annotation=Literal["single", "table"],
            default="single",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.rolling(**kwargs)


class pd_shift_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_SHIFT_GROUPBY"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="periods",
            annotation=int,
            default=1,
        ),
        SarusParameter(
            name="freq",
            annotation=Optional[pdt.Frequency],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="fill_value",
            annotation=Hashable,
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.shift(**kwargs)


class pd_groups_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_GROUPS_GROUPBY"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GroupBy,
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.groups


class pd_sum_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_SUM_GROUPBY"
    _dp_equivalent_id = "pandas.PD_SUM_GROUPBY_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GroupBy,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="min_count",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.sum(**kwargs)


class pd_std_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_STD_GROUPBY"
    _dp_equivalent_id = "pandas.PD_STD_GROUPBY_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GroupBy,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="ddof",
            annotation=int,
            default=1,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.std(**kwargs)


class pd_median_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_MEDIAN_GROUPBY"
    _dp_equivalent_id = "pandas.PD_MEDIAN_GROUPBY_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=GroupBy,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.median(**kwargs)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/transform.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import type_pb2 as sarus__data__spec_dot_protobuf_dot_type__pb2
from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2
from sarus_data_spec.protobuf import scalar_pb2 as sarus__data__spec_dot_protobuf_dot_scalar__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(sarus_data_spec/protobuf/transform.proto\x1a#sarus_data_spec/protobuf/type.proto\x1a#sarus_data_spec/protobuf/path.proto\x1a%sarus_data_spec/protobuf/scalar.proto\"\x8d#\n\tTransform\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12\x1d\n\x04spec\x18\x04 \x01(\x0b\x32\x0f.Transform.Spec\x12.\n\nproperties\x18\x05 \x03(\x0b\x32\x1a.Transform.PropertiesEntry\x12\x12\n\ninversible\x18\x06 \x01(\x08\x12\x19\n\x11schema_preserving\x18\x07 \x01(\x08\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xa3\x0c\n\x04Spec\x12\'\n\x08identity\x18\x01 \x01(\x0b\x32\x13.Transform.IdentityH\x00\x12\'\n\x08variable\x18\x02 \x01(\x0b\x32\x13.Transform.VariableH\x00\x12\'\n\x08\x63omposed\x18\x03 \x01(\x0b\x32\x13.Transform.ComposedH\x00\x12%\n\x07project\x18\x04 \x01(\x0b\x32\x12.Transform.ProjectH\x00\x12#\n\x06\x66ilter\x18\x05 \x01(\x0b\x32\x11.Transform.FilterH\x00\x12%\n\x07shuffle\x18\x06 \x01(\x0b\x32\x12.Transform.ShuffleH\x00\x12\x1f\n\x04join\x18\x07 \x01(\x0b\x32\x0f.Transform.JoinH\x00\x12\x1f\n\x04\x63\x61st\x18\x08 \x01(\x0b\x32\x0f.Transform.CastH\x00\x12#\n\x06sample\x18\t \x01(\x0b\x32\x11.Transform.SampleH\x00\x12\x30\n\ruser_settings\x18\n \x01(\x0b\x32\x17.Transform.UserSettingsH\x00\x12-\n\x0fprotect_dataset\x18\x0b \x01(\x0b\x32\x12.Transform.ProtectH\x00\x12\'\n\x08\x65xternal\x18\x0c \x01(\x0b\x32\x13.Transform.ExternalH\x00\x12)\n\tsynthetic\x18\r \x01(\x0b\x32\x14.Transform.SyntheticH\x00\x12)\n\ttranscode\x18\x0e \x01(\x0b\x32\x14.Transform.TranscodeH\x00\x12\x38\n\x11inverse_transcode\x18\x0f \x01(\x0b\x32\x1b.Transform.InverseTranscodeH\x00\x12&\n\x08get_item\x18\x10 \x01(\x0b\x32\x12.Transform.GetItemH\x00\x12\x34\n\x0fprotected_paths\x18\x11 \x01(\x0b\x32\x19.Transform.ProtectedPathsH\x00\x12\x43\n\x17\x61utomatic_user_settings\x18\x12 \x01(\x0b\x32 .Transform.AutomaticUserSettingsH\x00\x12.\n\x0cpublic_paths\x18\x13 \x01(\x0b\x32\x16.Transform.PublicPathsH\x00\x12\x30\n\rassign_budget\x18\x14 \x01(\x0b\x32\x17.Transform.AssignBudgetH\x00\x12\x36\n\x10\x61utomatic_budget\x18\x15 \x01(\x0b\x32\x1a.Transform.AutomaticBudgetH\x00\x12\x37\n\x10\x61ttribute_budget\x18\x16 \x01(\x0b\x32\x1b.Transform.AttributesBudgetH\x00\x12(\n\tsd_budget\x18\x17 \x01(\x0b\x32\x13.Transform.SDBudgetH\x00\x12,\n\x0b\x64\x65rive_seed\x18\x18 \x01(\x0b\x32\x15.Transform.DeriveSeedH\x00\x12+\n\x0bgroup_by_pe\x18\x19 \x01(\x0b\x32\x14.Transform.GroupByPEH\x00\x12\x34\n\x0fsampling_ratios\x18\x1a \x01(\x0b\x32\x19.Transform.SamplingRatiosH\x00\x12*\n\nselect_sql\x18\x1b \x01(\x0b\x32\x14.Transform.SelectSqlH\x00\x12%\n\x07\x65xtract\x18\x1c \x01(\x0b\x32\x12.Transform.ExtractH\x00\x12\x38\n\x11relationship_spec\x18\x1d \x01(\x0b\x32\x1b.Transform.RelationshipSpecH\x00\x12@\n\x15\x64ifferentiated_sample\x18\x1e \x01(\x0b\x32\x1f.Transform.DifferentiatedSampleH\x00\x12;\n\x13validated_user_type\x18\x1f \x01(\x0b\x32\x1c.Transform.ValidatedUserTypeH\x00\x12/\n\rdp_select_sql\x18  \x01(\x0b\x32\x16.Transform.DPSelectSqlH\x00\x12\x36\n\x10\x65rror_estimation\x18! \x01(\x0b\x32\x1a.Transform.ErrorEstimationH\x00\x42\x06\n\x04spec\x1a\xa4\x07\n\x08\x45xternal\x12\x11\n\targuments\x18\x01 \x01(\x0c\x12\x17\n\x0fnamed_arguments\x18\x02 \x01(\x0c\x12\x37\n\rop_identifier\x18\x03 \x01(\x0b\x32 .Transform.External.OpIdentifier\x1a\xa1\x04\n\x0cOpIdentifier\x12&\n\x03std\x18\x01 \x01(\x0b\x32\x17.Transform.External.StdH\x00\x12,\n\x06pandas\x18\x02 \x01(\x0b\x32\x1a.Transform.External.PandasH\x00\x12*\n\x05numpy\x18\x03 \x01(\x0b\x32\x19.Transform.External.NumpyH\x00\x12\x34\n\ntensorflow\x18\x04 \x01(\x0b\x32\x1e.Transform.External.TensorflowH\x00\x12.\n\x07sklearn\x18\x05 \x01(\x0b\x32\x1b.Transform.External.SklearnH\x00\x12?\n\x10pandas_profiling\x18\x06 \x01(\x0b\x32#.Transform.External.PandasProfilingH\x00\x12.\n\x07xgboost\x18\x07 \x01(\x0b\x32\x1b.Transform.External.XGBoostH\x00\x12*\n\x05skopt\x18\x08 \x01(\x0b\x32\x19.Transform.External.SkoptH\x00\x12\x30\n\x08imblearn\x18\t \x01(\x0b\x32\x1c.Transform.External.ImblearnH\x00\x12(\n\x04shap\x18\n \x01(\x0b\x32\x18.Transform.External.ShapH\x00\x12*\n\x05scipy\x18\x0b \x01(\x0b\x32\x19.Transform.External.ScipyH\x00\x42\x04\n\x02op\x1a\x13\n\x03Std\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x16\n\x06Pandas\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Numpy\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x1a\n\nTensorflow\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x17\n\x07Sklearn\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x1f\n\x0fPandasProfiling\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x17\n\x07XGBoost\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Skopt\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x18\n\x08Imblearn\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x14\n\x04Shap\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x15\n\x05Scipy\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\n\n\x08Identity\x1a*\n\x08Variable\x12\x10\n\x08position\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\x1a\xa9\x01\n\x08\x43omposed\x12\x11\n\ttransform\x18\x01 \x01(\t\x12\x11\n\targuments\x18\x02 \x03(\t\x12@\n\x0fnamed_arguments\x18\x03 \x03(\x0b\x32\'.Transform.Composed.NamedArgumentsEntry\x1a\x35\n\x13NamedArgumentsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x34\n\x07Project\x12)\n\nprojection\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a/\n\x06\x46ilter\x12%\n\x06\x66ilter\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a\t\n\x07Shuffle\x1a)\n\x04Join\x12!\n\x02on\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a+\n\x04\x43\x61st\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a\x61\n\x06Sample\x12\x12\n\x08\x66raction\x18\x01 \x01(\x01H\x00\x12\x0e\n\x04size\x18\x02 \x01(\x03H\x00\x12%\n\x04seed\x18\x03 \x01(\x0b\x32\x17.sarus_data_spec.ScalarB\x0c\n\nproportion\x1av\n\x0fSchemaInference\x12:\n\x0b\x63\x61st_policy\x18\x01 \x01(\x0e\x32%.Transform.SchemaInference.CastPolicy\"\'\n\nCastPolicy\x12\x08\n\x04NONE\x10\x00\x12\x0f\n\x0bMOST_LIKELY\x10\x01\x1a\x16\n\x07GroupBy\x12\x0b\n\x03key\x18\x01 \x01(\t\x1a\x0b\n\tSynthetic\x1a\x0e\n\x0cUserSettings\x1a/\n\x15\x41utomaticUserSettings\x12\x16\n\x0emax_categories\x18\x01 \x01(\x03\x1a\t\n\x07Protect\x1a\x0b\n\tTranscode\x1a\x12\n\x10InverseTranscode\x1ao\n\x14\x44ifferentiatedSample\x12\x12\n\x08\x66raction\x18\x01 \x01(\x01H\x00\x12\x0e\n\x04size\x18\x02 \x01(\x03H\x00\x12%\n\x04seed\x18\x03 \x01(\x0b\x32\x17.sarus_data_spec.ScalarB\x0c\n\nproportion\x1a\x10\n\x0eProtectedPaths\x1a\r\n\x0bPublicPaths\x1a.\n\x07GetItem\x12#\n\x04path\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\x1a\x0e\n\x0c\x41ssignBudget\x1a\x11\n\x0f\x41utomaticBudget\x1a\x12\n\x10\x41ttributesBudget\x1a\n\n\x08SDBudget\x1a$\n\nDeriveSeed\x12\x16\n\x0erandom_integer\x18\x01 \x01(\x03\x1a\x0b\n\tGroupByPE\x1a\x10\n\x0eSamplingRatios\x1a\x12\n\x10RelationshipSpec\x1a\x88\x01\n\tSelectSql\x12\x0f\n\x05query\x18\x01 \x01(\tH\x00\x12\x34\n\x0f\x61liased_queries\x18\x02 \x01(\x0b\x32\x19.Transform.AliasedQueriesH\x00\x12*\n\x0bsql_dialect\x18\x03 \x01(\x0e\x32\x15.Transform.SQLDialectB\x08\n\x06select\x1a\x8a\x01\n\x0b\x44PSelectSql\x12\x0f\n\x05query\x18\x01 \x01(\tH\x00\x12\x34\n\x0f\x61liased_queries\x18\x02 \x01(\x0b\x32\x19.Transform.AliasedQueriesH\x00\x12*\n\x0bsql_dialect\x18\x03 \x01(\x0e\x32\x15.Transform.SQLDialectB\x08\n\x06select\x1a@\n\x0e\x41liasedQueries\x12.\n\raliased_query\x18\x01 \x03(\x0b\x32\x17.Transform.AliasedQuery\x1a\x42\n\x0c\x41liasedQuery\x12#\n\x04path\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Path\x12\r\n\x05query\x18\x02 \x01(\t\x1a\x17\n\x07\x45xtract\x12\x0c\n\x04size\x18\x01 \x01(\x05\x1a\x13\n\x11ValidatedUserType\x1a\x11\n\x0f\x45rrorEstimation\"\x80\x01\n\nSQLDialect\x12\x08\n\x04NONE\x10\x00\x12\x0c\n\x08POSTGRES\x10\x01\x12\x0e\n\nSQL_SERVER\x10\x02\x12\n\n\x06MY_SQL\x10\x03\x12\x0b\n\x07SQLLITE\x10\x04\x12\n\n\x06ORACLE\x10\x05\x12\r\n\tBIG_QUERY\x10\x06\x12\x0c\n\x08REDSHIFT\x10\x07\x12\x08\n\x04HIVE\x10\x08\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.transform_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TRANSFORM_PROPERTIESENTRY._options = None
  _TRANSFORM_PROPERTIESENTRY._serialized_options = b'8\001'
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._options = None
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_options = b'8\001'
  _TRANSFORM._serialized_start=158
  _TRANSFORM._serialized_end=4651
  _TRANSFORM_PROPERTIESENTRY._serialized_start=338
  _TRANSFORM_PROPERTIESENTRY._serialized_end=387
  _TRANSFORM_SPEC._serialized_start=390
  _TRANSFORM_SPEC._serialized_end=1961
  _TRANSFORM_EXTERNAL._serialized_start=1964
  _TRANSFORM_EXTERNAL._serialized_end=2896
  _TRANSFORM_EXTERNAL_OPIDENTIFIER._serialized_start=2078
  _TRANSFORM_EXTERNAL_OPIDENTIFIER._serialized_end=2623
  _TRANSFORM_EXTERNAL_STD._serialized_start=2625
  _TRANSFORM_EXTERNAL_STD._serialized_end=2644
  _TRANSFORM_EXTERNAL_PANDAS._serialized_start=2646
  _TRANSFORM_EXTERNAL_PANDAS._serialized_end=2668
  _TRANSFORM_EXTERNAL_NUMPY._serialized_start=2670
  _TRANSFORM_EXTERNAL_NUMPY._serialized_end=2691
  _TRANSFORM_EXTERNAL_TENSORFLOW._serialized_start=2693
  _TRANSFORM_EXTERNAL_TENSORFLOW._serialized_end=2719
  _TRANSFORM_EXTERNAL_SKLEARN._serialized_start=2721
  _TRANSFORM_EXTERNAL_SKLEARN._serialized_end=2744
  _TRANSFORM_EXTERNAL_PANDASPROFILING._serialized_start=2746
  _TRANSFORM_EXTERNAL_PANDASPROFILING._serialized_end=2777
  _TRANSFORM_EXTERNAL_XGBOOST._serialized_start=2779
  _TRANSFORM_EXTERNAL_XGBOOST._serialized_end=2802
  _TRANSFORM_EXTERNAL_SKOPT._serialized_start=2804
  _TRANSFORM_EXTERNAL_SKOPT._serialized_end=2825
  _TRANSFORM_EXTERNAL_IMBLEARN._serialized_start=2827
  _TRANSFORM_EXTERNAL_IMBLEARN._serialized_end=2851
  _TRANSFORM_EXTERNAL_SHAP._serialized_start=2853
  _TRANSFORM_EXTERNAL_SHAP._serialized_end=2873
  _TRANSFORM_EXTERNAL_SCIPY._serialized_start=2875
  _TRANSFORM_EXTERNAL_SCIPY._serialized_end=2896
  _TRANSFORM_IDENTITY._serialized_start=2898
  _TRANSFORM_IDENTITY._serialized_end=2908
  _TRANSFORM_VARIABLE._serialized_start=2910
  _TRANSFORM_VARIABLE._serialized_end=2952
  _TRANSFORM_COMPOSED._serialized_start=2955
  _TRANSFORM_COMPOSED._serialized_end=3124
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_start=3071
  _TRANSFORM_COMPOSED_NAMEDARGUMENTSENTRY._serialized_end=3124
  _TRANSFORM_PROJECT._serialized_start=3126
  _TRANSFORM_PROJECT._serialized_end=3178
  _TRANSFORM_FILTER._serialized_start=3180
  _TRANSFORM_FILTER._serialized_end=3227
  _TRANSFORM_SHUFFLE._serialized_start=3229
  _TRANSFORM_SHUFFLE._serialized_end=3238
  _TRANSFORM_JOIN._serialized_start=3240
  _TRANSFORM_JOIN._serialized_end=3281
  _TRANSFORM_CAST._serialized_start=3283
  _TRANSFORM_CAST._serialized_end=3326
  _TRANSFORM_SAMPLE._serialized_start=3328
  _TRANSFORM_SAMPLE._serialized_end=3425
  _TRANSFORM_SCHEMAINFERENCE._serialized_start=3427
  _TRANSFORM_SCHEMAINFERENCE._serialized_end=3545
  _TRANSFORM_SCHEMAINFERENCE_CASTPOLICY._serialized_start=3506
  _TRANSFORM_SCHEMAINFERENCE_CASTPOLICY._serialized_end=3545
  _TRANSFORM_GROUPBY._serialized_start=3547
  _TRANSFORM_GROUPBY._serialized_end=3569
  _TRANSFORM_SYNTHETIC._serialized_start=3571
  _TRANSFORM_SYNTHETIC._serialized_end=3582
  _TRANSFORM_USERSETTINGS._serialized_start=3584
  _TRANSFORM_USERSETTINGS._serialized_end=3598
  _TRANSFORM_AUTOMATICUSERSETTINGS._serialized_start=3600
  _TRANSFORM_AUTOMATICUSERSETTINGS._serialized_end=3647
  _TRANSFORM_PROTECT._serialized_start=3649
  _TRANSFORM_PROTECT._serialized_end=3658
  _TRANSFORM_TRANSCODE._serialized_start=3660
  _TRANSFORM_TRANSCODE._serialized_end=3671
  _TRANSFORM_INVERSETRANSCODE._serialized_start=3673
  _TRANSFORM_INVERSETRANSCODE._serialized_end=3691
  _TRANSFORM_DIFFERENTIATEDSAMPLE._serialized_start=3693
  _TRANSFORM_DIFFERENTIATEDSAMPLE._serialized_end=3804
  _TRANSFORM_PROTECTEDPATHS._serialized_start=3806
  _TRANSFORM_PROTECTEDPATHS._serialized_end=3822
  _TRANSFORM_PUBLICPATHS._serialized_start=3824
  _TRANSFORM_PUBLICPATHS._serialized_end=3837
  _TRANSFORM_GETITEM._serialized_start=3839
  _TRANSFORM_GETITEM._serialized_end=3885
  _TRANSFORM_ASSIGNBUDGET._serialized_start=3887
  _TRANSFORM_ASSIGNBUDGET._serialized_end=3901
  _TRANSFORM_AUTOMATICBUDGET._serialized_start=3903
  _TRANSFORM_AUTOMATICBUDGET._serialized_end=3920
  _TRANSFORM_ATTRIBUTESBUDGET._serialized_start=3922
  _TRANSFORM_ATTRIBUTESBUDGET._serialized_end=3940
  _TRANSFORM_SDBUDGET._serialized_start=3942
  _TRANSFORM_SDBUDGET._serialized_end=3952
  _TRANSFORM_DERIVESEED._serialized_start=3954
  _TRANSFORM_DERIVESEED._serialized_end=3990
  _TRANSFORM_GROUPBYPE._serialized_start=3992
  _TRANSFORM_GROUPBYPE._serialized_end=4003
  _TRANSFORM_SAMPLINGRATIOS._serialized_start=4005
  _TRANSFORM_SAMPLINGRATIOS._serialized_end=4021
  _TRANSFORM_RELATIONSHIPSPEC._serialized_start=4023
  _TRANSFORM_RELATIONSHIPSPEC._serialized_end=4041
  _TRANSFORM_SELECTSQL._serialized_start=4044
  _TRANSFORM_SELECTSQL._serialized_end=4180
  _TRANSFORM_DPSELECTSQL._serialized_start=4183
  _TRANSFORM_DPSELECTSQL._serialized_end=4321
  _TRANSFORM_ALIASEDQUERIES._serialized_start=4323
  _TRANSFORM_ALIASEDQUERIES._serialized_end=4387
  _TRANSFORM_ALIASEDQUERY._serialized_start=4389
  _TRANSFORM_ALIASEDQUERY._serialized_end=4455
  _TRANSFORM_EXTRACT._serialized_start=4457
  _TRANSFORM_EXTRACT._serialized_end=4480
  _TRANSFORM_VALIDATEDUSERTYPE._serialized_start=4482
  _TRANSFORM_VALIDATEDUSERTYPE._serialized_end=4501
  _TRANSFORM_ERRORESTIMATION._serialized_start=4503
  _TRANSFORM_ERRORESTIMATION._serialized_end=4520
  _TRANSFORM_SQLDIALECT._serialized_start=4523
  _TRANSFORM_SQLDIALECT._serialized_end=4651
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import Collection, List, Optional, cast
import logging

from sarus_data_spec.constants import NO_TOKEN, PEP_TOKEN
from sarus_data_spec.dataspec_rewriter.utils import graph_private_queries
from sarus_data_spec.dataspec_validator.base import BaseDataspecValidator
from sarus_data_spec.dataspec_validator.privacy_limit import DeltaEpsilonLimit
from sarus_data_spec.manager.ops.processor import routing
from sarus_data_spec.storage.typing import Storage
from sarus_data_spec.variant_constraint import pep_for_rewriting_constraint
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)


class RecursiveDataspecValidator(BaseDataspecValidator):
    def __init__(self, storage: Storage):
        self._storage = storage

    # Local property
    def is_dp_able(self, dataspec: st.DataSpec) -> bool:
        """Checks if the dataspec has a transform that either has a DP
        equivalent, allowing the rewritten dataspec to be considered DP
        if the input rewritten PEP token is not None."""
        if not dataspec.is_transformed():
            return False

        if dataspec.is_dp():
            return True

        _, StaticChecker = routing.get_op(dataspec)
        return StaticChecker(dataspec).is_dp_able()

    def is_pep_able(self, dataspec: st.DataSpec) -> bool:
        """Checks if the dataspec has a transform that either has a PEP
        equivalent or does not require one, allowing the rewritten dataspec to
        be considered 'PEP' if the input rewritten PEP token is not None."""
        if dataspec.is_pep():
            return True

        _, StaticChecker = routing.get_op(dataspec)
        return StaticChecker(dataspec).is_pep_able()

    def is_dp_writable(self, dataspec: st.DataSpec) -> bool:
        if not self.is_dp_able(dataspec):
            return False

        if dataspec.is_dp():
            return True

        _, StaticChecker = routing.get_op(dataspec)
        return StaticChecker(dataspec).is_dp_writable([])

    def is_pep_writable(self, dataspec: st.DataSpec) -> bool:
        if not self.is_pep_able(dataspec):
            return False

        return self.rewritten_pep_token(dataspec) is not None

    def is_publishable(self, dataspec: st.DataSpec) -> bool:
        if not dataspec.is_transformed():
            return False

        if self.is_dp_writable(dataspec):
            return True

        parents = dataspec.parents_list()
        return all(
            [
                self.is_publishable(dataspec_parent)
                for dataspec_parent in parents
            ]
        )

    def is_published(self, dataspec: st.DataSpec) -> bool:
        """Check if the dataspec is the result of a DP transform or another
        published dataspec.
        There is at least one parent that is DP.
        Such a dataspec has at least one private query in its computation
        graph."""
        if not dataspec.is_transformed():
            return False

        if dataspec.is_dp():
            return True

        parents = dataspec.parents_list()
        return all(
            [self.is_published(dataspec_parent) for dataspec_parent in parents]
        )

    def rewritten_pep_token(self, dataspec: st.DataSpec) -> Optional[str]:
        """Returns the PEP token for the DP variant of this dataspec
        during DP rewriting.

        Currently, the implementation assumes a single DP/PEP variant per
        dataspec, resulting in one possible value for
        "rewritten_pep_token."
        Future changes could introduce multiple variants, necessitating
        the implementation of priority rules.
        """
        if dataspec.prototype() == sp.Scalar:
            return None

        dataset = cast(st.Dataset, dataspec)

        # TODO fetch real context and budget
        public_context: Collection[str] = []
        privacy_limit = DeltaEpsilonLimit({0.0: 0.0})
        kind = st.ConstraintKind.PEP_FOR_REWRITING

        for constraint in self.verified_constraints(dataspec):
            check_constraint = self.verifies(
                constraint, kind, public_context, privacy_limit
            )
            if check_constraint is not None:
                if check_constraint:
                    return constraint.properties()[PEP_TOKEN]
                else:
                    return None

        # Compute the PEP token
        if not dataset.is_transformed():
            return None

        transform = dataset.transform()
        _, StaticChecker = routing.get_dataset_op(transform)
        pep_token = StaticChecker(dataset).rewritten_pep_token(public_context)
        if pep_token is None:
            pep_token = NO_TOKEN

        pep_for_rewriting_constraint(
            dataspec=dataset,
            token=pep_token,
            required_context=[],
            privacy_limit=privacy_limit,
        )

        return None if pep_token == NO_TOKEN else pep_token

    def graph_private_queries(
        self, dataspec: st.DataSpec
    ) -> List[st.PrivateQuery]:
        """Return the list of PrivateQueries used in a Dataspec's transform.

        This method collect all the private queries of the  computation graph.

        It represents the privacy loss associated with the current computation.

        It can be used by Sarus when a user (Access object) reads a DP dataspec
        to update its accountant. Note that Private Query objects are generated
        with a random uuid so that even if they are submitted multiple times to
        an account, they are only accounted once (ask @cgastaud for more on
        accounting).
        """
        return graph_private_queries(dataspec)



from datetime import date, datetime, time, timedelta, timezone
from typing import Any
import json
import sys

from dateutil.parser import parse
import numpy as np
import pandas as pd


class SarusJSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, np.ndarray):
            return {'_type': 'numpy.ndarray', 'data': obj.tolist()}
        elif isinstance(obj, pd.DataFrame):
            return {
                '_type': 'pandas.DataFrame',
                'data': obj.to_json(date_format='iso'),
            }
        elif isinstance(obj, pd.Series):
            return {
                '_type': 'pandas.Series',
                'data': obj.to_json(date_format='iso'),
            }
        elif isinstance(obj, pd.Timestamp):
            return {'_type': 'pandas.Timestamp', 'data': obj.isoformat()}
        elif isinstance(obj, datetime):
            return {'_type': 'datetime', 'data': obj.isoformat()}
        elif isinstance(obj, timedelta):
            return {'_type': 'timedelta', 'data': obj.total_seconds()}
        elif isinstance(obj, timezone):
            utcoffset_result = obj.utcoffset(None)
            if utcoffset_result is not None:
                return {
                    '_type': 'timezone',
                    'data': utcoffset_result.total_seconds(),
                }
            else:
                raise ValueError("Invalid timezone object")
        elif isinstance(obj, time):
            return {'_type': 'time', 'data': obj.isoformat()}
        elif isinstance(obj, date):
            return {'_type': 'date', 'data': obj.isoformat()}
        elif isinstance(obj, np.generic):
            if np.issubdtype(obj, np.complexfloating):
                complex_obj = obj.astype(np.complex128)
                return {
                    '_type': 'numpy.complex',
                    'data': {
                        '"real"': complex_obj.real,
                        '"imag"': complex_obj.imag,
                    },
                }
            else:
                return {'_type': 'numpy.generic', 'data': obj.item()}
        elif isinstance(obj, pd.MultiIndex):
            return {
                '_type': 'pandas.MultiIndex',
                'data': obj.tolist(),
                'names': obj.names,
                'levels': [level.tolist() for level in obj.levels],
                'codes': list(obj.codes),
            }
        elif isinstance(obj, pd.Index):
            return {
                '_type': 'pandas.Index',
                'class': type(obj).__name__,
                'data': obj.tolist(),
                'dtype': str(obj.dtype),
            }
        elif isinstance(obj, pd.Period):
            return {
                '_type': 'pandas.Period',
                'data': str(obj),
                'freq': obj.freqstr,
            }
        elif isinstance(obj, pd.Timedelta):
            return {'_type': 'pandas.Timedelta', 'data': obj.value}
        elif isinstance(obj, pd.Interval):
            return {'_type': 'pandas.Interval', 'data': (obj.left, obj.right)}
        elif isinstance(obj, pd.Categorical):
            return {
                '_type': 'pandas.Categorical',
                'data': obj.tolist(),
                'categories': obj.categories.tolist(),
                'ordered': obj.ordered,
            }
        elif isinstance(obj, type):
            return {
                '_type': 'class',
                'data': {'"name"': obj.__name__, '"module"': obj.__module__},
            }
        elif isinstance(obj, pd.api.extensions.ExtensionDtype):
            return {'_type': 'dtype', 'data': str(obj)}
        elif isinstance(obj, slice):
            return {'_type': 'slice', 'data': (obj.start, obj.stop, obj.step)}
        elif isinstance(obj, range):
            return {
                '_type': 'range',
                'data': {
                    '"start"': obj.start,
                    '"stop"': obj.stop,
                    '"step"': obj.step,
                },
            }
        return super().default(obj)

    def encode_obj(self, obj: Any) -> Any:
        if isinstance(obj, tuple):
            return {
                '_type': 'tuple',
                'data': [self.encode_obj(v) for v in obj],
            }
        elif isinstance(obj, list):
            return [self.encode_obj(v) for v in obj]
        elif isinstance(obj, dict):
            return {self.encode(k): self.encode_obj(v) for k, v in obj.items()}
        return obj

    def encode(self, obj: Any) -> str:
        obj_transformed = self.encode_obj(obj)
        return super().encode(obj_transformed)

    @classmethod
    def encode_bytes(cls, obj: Any) -> bytes:
        encoder = cls()
        return (encoder.encode(obj)).encode('utf-8')


class SarusJSONDecoder(json.JSONDecoder):
    def decode(self, s: str, *args: Any, **kwargs: Any) -> Any:
        obj = super().decode(s, *args, **kwargs)
        return self.decode_obj(obj)

    def decode_obj(self, obj: Any) -> Any:
        if isinstance(obj, dict):
            if '_type' in obj:
                data = self.decode_obj(obj['data'])
                if obj['_type'] == 'tuple':
                    return tuple(self.decode_obj(v) for v in data)
                elif obj['_type'] == 'numpy.ndarray':
                    return np.array(data)
                elif obj['_type'] == 'pandas.DataFrame':
                    return pd.read_json(data, convert_dates=True)
                elif obj['_type'] == 'pandas.Series':
                    return pd.read_json(data, typ='series', convert_dates=True)
                elif obj['_type'] == 'pandas.Timestamp':
                    return pd.Timestamp(data)
                elif obj['_type'] == 'datetime':
                    return parse(data)
                elif obj['_type'] == 'timedelta':
                    return timedelta(seconds=data)
                elif obj['_type'] == 'timezone':
                    return timezone(timedelta(seconds=data))
                elif obj['_type'] == 'time':
                    return parse(data).time()
                elif obj['_type'] == 'date':
                    return parse(data).date()
                elif obj['_type'] == 'numpy.generic':
                    return np.array(data).item()
                elif obj['_type'] == 'numpy.complex':
                    return np.complex128(complex(data['real'], data['imag']))
                elif obj['_type'] == 'pandas.Index':
                    cls = getattr(pd, obj['class'])
                    return cls(data, dtype=obj['dtype'])
                elif obj['_type'] == 'pandas.MultiIndex':
                    return pd.MultiIndex(
                        levels=[pd.Index(level) for level in obj['levels']],
                        codes=self.decode_obj(obj['codes']),
                        names=obj['names'],
                    )
                elif obj['_type'] == 'pandas.Period':
                    return pd.Period(data, freq=obj['freq'])
                elif obj['_type'] == 'pandas.Timedelta':
                    return pd.to_timedelta(data)
                elif obj['_type'] == 'pandas.Interval':
                    return pd.Interval(*data)
                elif obj['_type'] == 'pandas.Categorical':
                    return pd.Categorical(
                        data,
                        categories=obj['categories'],
                        ordered=obj['ordered'],
                    )
                elif obj['_type'] == 'class':
                    if data['module'] in ['builtins', 'numpy', 'pandas']:
                        cls = getattr(
                            sys.modules[data['module']], data['name']
                        )
                        if isinstance(cls, type):
                            return cls
                        else:
                            raise ValueError("Decoded object is not a type")
                    else:
                        raise ValueError("Invalid module name")
                elif obj['_type'] == 'dtype':
                    return pd.api.types.pandas_dtype(data)
                elif obj['_type'] == 'slice':
                    return slice(*data)
                elif obj['_type'] == 'range':
                    return range(data['start'], data['stop'], data['step'])
            return {self.decode(k): self.decode_obj(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self.decode_obj(v) for v in obj]
        return obj

    @classmethod
    def decode_bytes(cls, b: bytes, *args: Any, **kwargs: Any) -> Any:
        decoder = cls()
        return decoder.decode(b.decode('utf-8'), *args, **kwargs)



from time import time_ns
from typing import Any, Collection, Dict, List, Optional, Tuple, cast
import logging

from sarus_data_spec.context import global_context
from sarus_data_spec.dataset import transformed
from sarus_data_spec.dataspec_rewriter.base import BaseDataspecRewriter
from sarus_data_spec.dataspec_rewriter.graph_query_builder_utils import (
    GraphBuilder,
    build_and_fit_graph_builder,
    extract_privacy_limit,
)
from sarus_data_spec.dataspec_rewriter.simple_rules import (
    attach_variant,
    rewrite_mock,
    rewrite_synthetic,
)
from sarus_data_spec.dataspec_rewriter.utils import (
    find_dataspec_from_constraint,
)
from sarus_data_spec.dataspec_validator.privacy_limit import DeltaEpsilonLimit
from sarus_data_spec.dataspec_validator.recursive_validator import (
    RecursiveDataspecValidator,
)
from sarus_data_spec.manager.ops.processor import routing
from sarus_data_spec.scalar import privacy_budget
from sarus_data_spec.storage.typing import Storage
from sarus_data_spec.variant_constraint import dp_constraint
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

try:
    from sarus_query_builder.builders.composed_builder import ComposedBuilder
except ModuleNotFoundError as e_pandas_dp:
    if "sarus_query_builder" not in str(e_pandas_dp):
        raise
    ComposedBuilder = Any  # type: ignore
    OptimizableQueryBuilder = Any

logger = logging.getLogger(__name__)


class RecursiveDataspecRewriter(BaseDataspecRewriter):
    def __init__(self, storage: Storage):
        self._storage = storage
        self._dataspec_validator = RecursiveDataspecValidator(storage=storage)
        self._rewriter_dict: Dict[
            Tuple[str, str], str
        ] = (
            {}
        )  # help to rewrite, avoid rewriting multipletime the same dataspec

    def rewriter_dict(self) -> Dict[Tuple[str, str], str]:
        return self._rewriter_dict

    def rewrite_uuid(
        self,
        dataspec_uuid: str,
        salt: Optional[int] = None,
        privacy_limit: Optional[st.PrivacyLimit] = None,
    ) -> str:
        if salt is None:
            salt = time_ns()
        rewrite_uuid = dataspec_uuid + str(salt)
        return rewrite_uuid

    def get_transform_for_dp_rewriting(
        self, dataspec: st.DataSpec
    ) -> st.Transform:
        """Returns the appropriate transform for rewriting
        a dataspec in the context of DP rewriting."""
        if not dataspec.is_transformed():
            raise ValueError(
                """A not transformed dataspec does not
                have transform for rewriting"""
            )

        if dataspec.is_dp_writable():
            _, StaticChecker = routing.get_op(dataspec)
            transform_to_use = StaticChecker(dataspec).dp_transform()
            if transform_to_use is None:
                raise ValueError(
                    "A DP applicable dataspec should have a dp transform"
                )
        else:
            transform_to_use = dataspec.transform()
        return transform_to_use

    def get_transform_for_pep_rewriting(
        self, dataspec: st.DataSpec
    ) -> st.Transform:
        """Returns the appropriate transform for rewriting
        a dataspec in the context of PEP rewriting."""
        if not dataspec.is_transformed():
            raise ValueError(
                """A not transformed dataspec does not
                have transform for rewriting"""
            )

        if dataspec.is_pep_writable():
            dataset = cast(st.Dataset, dataspec)
            _, StaticChecker = routing.get_op(dataspec)
            transform_to_use = StaticChecker(dataset).pep_transform()
            if transform_to_use is None:
                raise ValueError(
                    "A PEP applicable dataspec should have a pep transform"
                )
        else:
            transform_to_use = dataspec.transform()
        return transform_to_use

    def variant(
        self,
        dataspec: st.DataSpec,
        kind: st.ConstraintKind,
        public_context: Collection[str],
        privacy_limit: Optional[st.PrivacyLimit],
        salt: Optional[int] = None,
    ) -> Optional[st.DataSpec]:
        return self.rewrite(
            dataspec,
            kind,
            public_context,
            privacy_limit,
            salt,
        )

    def rewrite(
        self,
        dataspec: st.DataSpec,
        target_kind: st.ConstraintKind,
        public_context: Collection[str],
        privacy_limit: Optional[st.PrivacyLimit],
        salt: Optional[int] = None,
    ) -> Optional[st.DataSpec]:
        """Returns a compliant dataspec or None."""

        if target_kind == st.ConstraintKind.SYNTHETIC:
            variant, _ = rewrite_synthetic(
                self.dataspec_validator(),
                dataspec,
                public_context,
            )
            return variant

        elif target_kind == st.ConstraintKind.MOCK:
            mock_variant, _ = rewrite_mock(
                self.dataspec_validator(),
                dataspec,
                public_context,
                salt,
            )
            return mock_variant

        if privacy_limit is None:
            raise ValueError(
                "Privacy limit must be defined for PEP or DP rewriting"
            )

        rewrite_uuid = self.rewrite_uuid(
            dataspec.uuid(), salt=salt, privacy_limit=privacy_limit
        )

        if target_kind == st.ConstraintKind.DP:
            if not dataspec.is_publishable():
                return rewrite_synthetic(
                    self.dataspec_validator(), dataspec, public_context
                )[0]
            # a dp rewriting require some minimum budget
            delta_epsilon_dict = privacy_limit.delta_epsilon_dict()
            _, epsilon = next(iter(delta_epsilon_dict.items()))
            if epsilon == 0:
                return rewrite_synthetic(
                    self.dataspec_validator(), dataspec, public_context
                )[0]
            graph_query_builder = build_and_fit_graph_builder(
                dataspec, privacy_limit
            )

            variant, _ = self.rewrite_dp(
                dataspec,
                public_context=public_context,
                graph_query_builder=graph_query_builder,
                rewrite_uuid=rewrite_uuid,
            )
            return variant

        elif target_kind == st.ConstraintKind.PEP:
            if not dataspec.is_pep_writable():
                return rewrite_synthetic(
                    self.dataspec_validator(), dataspec, public_context
                )[0]
            graph_query_builder = build_and_fit_graph_builder(
                dataspec, privacy_limit
            )
            variant, _ = self.rewrite_pep(
                dataspec,
                public_context=public_context,
                graph_query_builder=graph_query_builder,
                rewrite_uuid=rewrite_uuid,
            )
            return variant
        else:
            raise ValueError(
                f"Privacy policy {target_kind} rewriting not implemented yet"
            )

    def find_variant_from_rewriter(
        self, uuid: str, rewrite_uuid: Optional[str] = None
    ) -> Optional[st.DataSpec]:
        """Returna an already computed variant of a dataspec from the
        current rewriting defined by the rewrite uuid.

        Args:
            uuid (str): The uuid of the dataspec to be rewritten.
            rewrite_uuid (Optional[str], optional): The rewrite uuid to keep
            track of the already rewritten dataspec. Defaults to None.

        Returns:
            Optional[st.DataSpec]: The already computed variant or None.
        """
        if rewrite_uuid is None:
            return None
        # find already computed variant during the rewriting
        if (uuid, rewrite_uuid) in self.rewriter_dict():
            variant_uuid = self.rewriter_dict()[(uuid, rewrite_uuid)]
            variant = self.storage().referrable(variant_uuid)
            assert variant is not None
            dataspec = cast(st.DataSpec, variant)
            return dataspec
        else:
            return None

    def _rewrite_arg(
        self,
        arg: st.DataSpec,
        public_context: Collection[str],
        graph_query_builder: GraphBuilder,
        salt: Optional[int] = None,
        rewrite_uuid: Optional[str] = None,
    ) -> st.DataSpec:
        """
        Rewrites the dataspec 'arg' to the appropriate variant based on its
        properties.

        Currently, the function assumes that there is only one DP/PEP variant
        for any dataspec (excluding synthetic variants).
        This assumption may change in the future, and the function should be
        updated accordingly.

        Args:
            arg (st.DataSpec): The dataspec to be rewritten.
            public_context (Collection[str]): The public context for the
            rewriting process.
            graph_query_builder (GraphBuilder): The graph query builder for
            the dataspec being rewritten.
            salt (Optional[int], optional): An optional salt value for the
            rewriting process. Defaults to None.
            rewrite_uuid (Optional[str], optional): An optional UUID string
            for tracking the rewrite process. Defaults to None.

        Raises:
            ValueError: If the dataspec is not rewritable.

        Returns:
            st.DataSpec: The rewritten dataspec variant.
        """
        if isinstance(arg, st.DataSpec):
            if arg.is_publishable():
                rewritten_arg = self.rewrite_dp(
                    arg,
                    public_context,
                    graph_query_builder=graph_query_builder,
                    salt=salt,
                    rewrite_uuid=rewrite_uuid,
                )[0]
            elif arg.is_pep_writable():
                rewritten_arg = self.rewrite_pep(
                    arg,
                    public_context,
                    graph_query_builder=graph_query_builder,
                    salt=salt,
                    rewrite_uuid=rewrite_uuid,
                )[0]
            elif arg.is_public():
                rewritten_arg = arg
            else:
                raise ValueError(
                    f"""No rewriting possible for
                    the dataspec: {arg.uuid()}. """
                )
        else:
            rewritten_arg = arg
        return rewritten_arg

    def _rewrite_parents(
        self,
        dataspec: st.DataSpec,
        public_context: Collection[str],
        graph_query_builder: GraphBuilder,
        salt: Optional[int] = None,
        rewrite_uuid: Optional[str] = None,
    ) -> Tuple[List[st.DataSpec[Any]], Dict[str, st.DataSpec[Any]]]:
        """Rewrites all the parents of the dataspec to the appropriate variant
        based on its
        properties.

        Args:
            dataspec (st.DataSpec): The dataspec to be rewritten.
            public_context (Collection[str]): The public context for the
            rewriting process.
            graph_query_builder (GraphBuilder): The graph query builder for
            the dataspec being rewritten.
            salt (Optional[int], optional): An optional salt value for the
            rewriting process. Defaults to None.
            rewrite_uuid (Optional[str], optional): An optional UUID string
            for tracking the rewrite process. Defaults to None.

        Returns:
            Tuple[List[st.DataSpec[Any]], Dict[str, st.DataSpec[Any]]]:
            rewrited parents
        """
        args, kwargs = dataspec.parents()
        rewritten_args = []
        for arg in args:
            assert isinstance(arg, st.DataSpec)
            rewritten_arg = self._rewrite_arg(
                arg,
                public_context,
                graph_query_builder,
                salt,
                rewrite_uuid,
            )
            rewritten_args.append(rewritten_arg)

        rewritten_kwargs = {}
        for name, arg in kwargs.items():
            assert isinstance(arg, st.DataSpec)
            rewritten_arg = self._rewrite_arg(
                arg,
                public_context,
                graph_query_builder,
                salt,
                rewrite_uuid,
            )
            rewritten_kwargs[name] = rewritten_arg
        return rewritten_args, rewritten_kwargs

    def rewrite_dp(
        self,
        dataspec: st.DataSpec,
        public_context: Collection[str],
        graph_query_builder: GraphBuilder,
        salt: Optional[int] = None,
        rewrite_uuid: Optional[str] = None,
    ) -> Tuple[st.DataSpec, Collection[str]]:
        """
        Return an "is_publishable" dataspec into an "is_published" variant.

        Args:
            dataspec (st.DataSpec): The dataspec to be rewritten.
            public_context (Collection[str]): Not used yet.
            graph_query_builder (GraphBuilder): The graph query builder of the
            dataspec we are rewriting.
            salt (Optional[int]): A salt value for the rewriting, if required.
            Default is None.
            rewrite_uuid (Optional[str]): A UUID string for tracking the
            rewrite process. Default is None.

        Returns:
            Tuple[st.DataSpec, Collection[str]]: A tuple containing the
            "is_published" dataspec and an updated collection of public
            context information.

        Note:
            The dataspec provided must be "is_publishable" for the
            rewriting to proceed. Ensure that the dataspec
            meets this criterion before invoking this function.
        """
        if salt is None:
            salt = time_ns()

        if not dataspec.is_publishable():
            raise ValueError(
                "Try to rewrite with dp a non publishable dataspec"
            )

        if dataspec.is_published():
            return dataspec, public_context

        if graph_query_builder is None:
            raise ValueError(
                """A Dataspec dp publishable but not published should have a
                graph query builder to be published. """
            )

        # find already computed variant during the rewriting
        existing_variant = self.find_variant_from_rewriter(
            dataspec.uuid(), rewrite_uuid
        )
        if existing_variant is not None:
            return existing_variant, public_context

        # find already computed variant from constraint
        privacy_limit = extract_privacy_limit(graph_query_builder)
        if privacy_limit is None:
            raise ValueError(
                """The graph_query_builder is not fitted, thus no privacy
                limit is defined for the dp rewriting"""
            )

        kind = st.ConstraintKind.DP
        existing_variant = find_dataspec_from_constraint(
            dataspec_validator=self.dataspec_validator(),
            dataspec=dataspec,
            kind=kind,
            public_context=public_context,
            privacy_limit=privacy_limit,
            salt=salt,
        )
        if existing_variant is not None:
            return existing_variant, public_context

        # rewrite parents
        rewritten_args, rewritten_kwargs = self._rewrite_parents(
            dataspec=dataspec,
            public_context=public_context,
            graph_query_builder=graph_query_builder,
            rewrite_uuid=rewrite_uuid,
        )

        # create dataspec
        transform_to_use = self.get_transform_for_dp_rewriting(
            dataspec=dataspec
        )
        if dataspec.is_dp_writable():
            dict_builder = (
                graph_query_builder.split_graph_builder_in_dict_of_builders()
            )
            if dataspec.uuid() not in dict_builder:
                raise ValueError(
                    "A is_dp_writable dataspec should have a builder"
                )
            builder = dict_builder[dataspec.uuid()]
            dataspec_epsilon_deltas_budget = builder.epsilon_deltas_budget
            if dataspec_epsilon_deltas_budget is None:
                raise ValueError(
                    """The graph_query_builder is not fitted, thus no budget
                 is defined for the dp rewriting"""
                )
            epsilon, delta = list(dataspec_epsilon_deltas_budget)[0]
            dataspec_privacy_limit = DeltaEpsilonLimit({delta: epsilon})

            budget = privacy_budget(dataspec_privacy_limit)
            if salt is None:
                raise ValueError("A salt must be defined for rewriting in DP")
            seed = global_context().generate_seed(salt=salt)

            dp_variant = cast(
                st.DataSpec,
                transformed(
                    transform_to_use,
                    *rewritten_args,
                    dataspec_type=sp.type_name(dataspec.prototype()),
                    dataspec_name=None,
                    budget=budget,
                    seed=seed,
                    **rewritten_kwargs,
                ),
            )
        else:
            dp_variant = cast(
                st.DataSpec,
                transformed(
                    transform_to_use,
                    *rewritten_args,
                    dataspec_type=sp.type_name(dataspec.prototype()),
                    dataspec_name=None,
                    **rewritten_kwargs,
                ),
            )
        # cache result in rewriter
        if rewrite_uuid is not None:
            self.rewriter_dict()[
                (dataspec.uuid(), rewrite_uuid)
            ] = dp_variant.uuid()

        # attach constraints
        dp_constraint(
            dataspec=dp_variant,
            required_context=list(public_context),
            privacy_limit=privacy_limit,
            salt=salt,
        )
        attach_variant(
            original=dataspec,
            variant=dp_variant,
            kind=st.ConstraintKind.DP,
        )

        # We also attach the dataspec's synthetic/mock variant
        # to be the DP dataspec's synthetic variant. This is to
        # avoid to have DP computations in the MOCK.
        syn_variant = dataspec.variant(st.ConstraintKind.SYNTHETIC)
        if syn_variant is None:
            raise ValueError("Could not find a synthetic variant.")
        attach_variant(
            original=dp_variant,
            variant=syn_variant,
            kind=st.ConstraintKind.SYNTHETIC,
        )

        mock_variant = dataspec.variant(st.ConstraintKind.MOCK)
        if mock_variant is None:
            raise ValueError("Could not find a mock variant.")
        attach_variant(
            original=dp_variant,
            variant=mock_variant,
            kind=st.ConstraintKind.MOCK,
        )

        assert dp_variant.is_published()
        return dp_variant, public_context

    def rewrite_pep(
        self,
        dataspec: st.DataSpec,
        public_context: Collection[str],
        graph_query_builder: GraphBuilder,
        salt: Optional[int] = None,
        rewrite_uuid: Optional[str] = None,
    ) -> Tuple[st.DataSpec, Collection[str]]:
        """
        Return an "is_pep_writable" dataspec into an pep variant.

        Args:
            dataspec (st.DataSpec): The dataspec to be rewritten.
            public_context (Collection[str]): Not used yet.
            graph_query_builder (GraphBuilder): The graph query builder of the
            dataspec we are rewriting
            salt (Optional[int]): A salt value for the rewriting, if required.
            Default is None.
            rewrite_uuid (Optional[str]): A UUID string for tracking the
            rewrite process. Default is None.

        Returns:
            Tuple[st.DataSpec, Collection[str]]: A tuple containing the
            "is_pep" dataspec and an updated collection of public
            context information.

        Note:
            The dataspec provided must be "is_pep_writable" for the
            rewriting to proceed. Ensure that the dataspec
            meets this criterion before invoking this function.
        """
        if not dataspec.is_pep_writable():
            raise ValueError(
                "Try to rewrite in pep a non pep appicable dataspec"
            )

        if dataspec.is_pep():
            return dataspec, public_context

        # find already computed variant during the rewriting
        existing_variant = self.find_variant_from_rewriter(
            dataspec.uuid(), rewrite_uuid
        )
        if existing_variant is not None:
            return existing_variant, public_context

        if graph_query_builder is not None:
            privacy_limit = extract_privacy_limit(graph_query_builder)

        # find already computed variant from constraint
        kind = st.ConstraintKind.PEP
        variant = find_dataspec_from_constraint(
            dataspec_validator=self.dataspec_validator(),
            dataspec=dataspec,
            kind=kind,
            public_context=public_context,
            privacy_limit=privacy_limit,
            salt=salt,
        )
        if variant is not None:
            return variant, public_context

        # rewrite parents
        rewritten_args, rewritten_kwargs = self._rewrite_parents(
            dataspec=dataspec,
            public_context=public_context,
            graph_query_builder=graph_query_builder,
            rewrite_uuid=rewrite_uuid,
        )

        # create dataspec
        transform_to_use = self.get_transform_for_pep_rewriting(
            dataspec=dataspec
        )
        pep_variant = cast(
            st.DataSpec,
            transformed(
                transform_to_use,
                *rewritten_args,
                dataspec_type=sp.type_name(dataspec.prototype()),
                dataspec_name=None,
                **rewritten_kwargs,
            ),
        )

        # cache result in rewriter
        if rewrite_uuid is not None:
            self.rewriter_dict()[
                (dataspec.uuid(), rewrite_uuid)
            ] = pep_variant.uuid()

        # We also attach the dataspec's synthetic/mock variant
        # to be the DP dataspec's synthetic variant. This is to
        # avoid to have DP computations in the MOCK.
        syn_variant = dataspec.variant(st.ConstraintKind.SYNTHETIC)
        if syn_variant is None:
            raise ValueError("Could not find a synthetic variant.")
        attach_variant(
            original=pep_variant,
            variant=syn_variant,
            kind=st.ConstraintKind.SYNTHETIC,
        )

        mock_variant = dataspec.variant(st.ConstraintKind.MOCK)
        if mock_variant is None:
            raise ValueError("Could not find a mock variant.")
        attach_variant(
            original=pep_variant,
            variant=mock_variant,
            kind=st.ConstraintKind.MOCK,
        )

        # assert pep_variant.pep_token() is not None
        assert pep_variant.is_pep()
        return pep_variant, public_context



import typing as t

import pandas as pd

from sarus_data_spec.manager.ops.processor.standard.standard_op import (  # noqa: E501
    StandardScalarImplementation,
    StandardScalarStaticChecker,
)
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class ErrorEstimationStaticChecker(StandardScalarStaticChecker):
    ...


class ErrorEstimation(StandardScalarImplementation):
    """Computes the budget via standard rule
    depending on the number of columns of the
    parent dataspec"""

    async def value(self) -> t.Any:
        dataspecs = self.parents()

        if dataspecs[0].prototype() == sp.Scalar:
            true_value = await t.cast(st.Scalar, dataspecs[0]).async_value()
            dp_values = []
            for dataspec in dataspecs[1:]:
                dp_value = await t.cast(st.Scalar, dataspec).async_value()
                dp_values.append(abs(dp_value - true_value))

            return max(dp_values)
        else:
            true_value = t.cast(
                pd.Series,
                await dataspecs[0]
                .manager()
                .async_to(t.cast(st.Dataset, dataspecs[0]), pd.Series),
            )

            dp_values = []
            for dataspec in dataspecs[1:]:
                dp_value = t.cast(
                    pd.Series,
                    await dataspec.manager().async_to(
                        t.cast(st.Dataset, dataspec), pd.Series
                    ),
                )
                dp_values.append((dp_value - true_value).abs())
            results = pd.concat(dp_values, axis=1)
            return results.max(axis=1)
